# Licensed under a 3-clause BSD style license - see LICENSE.rst

import functools

import numpy as np
import pytest

from astropy import units as u
from astropy.coordinates import EarthLocation
from astropy.table import Table
from astropy.time import Time, conf
from astropy.utils import iers
from astropy.utils.compat import NUMPY_LT_1_25, NUMPY_LT_1_26
from astropy.utils.compat.optional_deps import HAS_H5PY
from astropy.utils.masked import Masked

allclose_sec = functools.partial(
    np.allclose, rtol=2.0**-52, atol=2.0**-52 * 24 * 3600
)  # 20 ps atol
is_masked = np.ma.is_masked


def test_simple():
    t = Time([1, 2, 3], format="cxcsec")
    assert t.masked is False
    assert np.all(t.mask == [False, False, False])

    # Before masking, format output does not have a mask
    # (it is an ndarray like always)
    assert not hasattr(t.value, "mask")
    assert not hasattr(t.unix, "mask")

    t[2] = np.ma.masked
    assert t.masked is True
    assert np.all(t.mask == [False, False, True])
    assert allclose_sec(t.value[:2], [1, 2])
    assert is_masked(t.value[2])
    assert is_masked(t[2].value)

    # After masking format output has a mask.
    assert hasattr(t.value, "mask")
    assert hasattr(t.unix, "mask")

    # Can also unmask.
    t[2] = np.ma.nomask
    assert np.all(t.mask == [False, False, False])
    # But since the internal data are still masked, the instance stays masked too,
    # as does any output.
    assert t.masked
    assert hasattr(t.value, "mask")
    assert hasattr(t.unix, "mask")
    # Combo just for completeness
    t[1:] = np.ma.masked
    t[1] = np.ma.nomask
    assert np.all(t.mask == [False, False, True])
    assert t.masked


def test_unmasked():
    t = Time([1, 2, 3], format="cxcsec")
    t[2] = np.ma.masked
    assert t.masked
    t_unmasked = t.unmasked
    assert not t_unmasked.masked
    assert not hasattr(t_unmasked.value, "mask")
    assert not hasattr(t_unmasked.unix, "mask")
    assert (t_unmasked.value == t.value.unmasked).all()
    # Check that data is shared.
    assert np.may_share_memory(t_unmasked._time.jd1, t._time.jd1)


@pytest.mark.parametrize("fill_value", [4, Time(5, format="cxcsec")])
def test_filled(fill_value):
    t = Time([1, 2, 3], format="cxcsec")
    t[2] = np.ma.masked
    # Fill with something suitable.
    t_filled = t.filled(fill_value)
    assert not t_filled.masked
    assert not hasattr(t_filled.value, "mask")
    assert not hasattr(t_filled.unix, "mask")
    expected = t.value.filled(Time(fill_value, format="cxcsec").value)
    assert (t_filled.value == expected).all()
    # Check that data is not shared.
    assert not np.may_share_memory(t_filled._time.jd1, t._time.jd1)


def test_scalar_init():
    t = Time("2000:001")
    assert t.masked is False
    assert t.mask == np.array(False)


def test_mask_not_writeable():
    t = Time("2000:001")
    with pytest.raises(
        AttributeError,
        match=rf"property 'mask' of '{t.__class__.__name__}' object has no setter",
    ):
        t.mask = True

    t = Time(["2000:001"])
    with pytest.raises(ValueError, match="assignment destination is read-only"):
        t.mask[0] = True

    # But we can set it to masked directly.
    t[0] = np.ma.masked
    assert np.all(t.mask == [True])
    # After this, the mask should again not be writeable.
    with pytest.raises(ValueError, match="assignment destination is read-only"):
        t.mask[0] = False

    # Should also not be writeable if we initialize with masked elements.
    t = Time(Masked(["2000:001", "2000:002"], mask=[False, True]))
    with pytest.raises(ValueError, match="assignment destination is read-only"):
        t.mask[0] = True
    # But again we can set it to masked directly.
    t[0] = np.ma.masked
    assert np.all(t.mask == [True, True])
    # Check that the mask remains shared.
    assert np.may_share_memory(t._time.jd1.mask, t._time.jd2.mask)


def test_str():
    t = Time(["2000:001", "2000:002"])
    t[1] = np.ma.masked
    assert str(t) == "['2000:001:00:00:00.000'                     ———]"
    assert (
        repr(t)
        == "<Time object: scale='utc' format='yday' value=['2000:001:00:00:00.000'                     ———]>"
    )

    expected = [
        "MaskedNDArray(['2000-01-01 00:00:00.000',                       ———],",
        "              dtype='<U23')",
    ]

    # Note that we need to take care to allow for big-endian platforms,
    # for which the dtype will be >U23 instead of <U23, which we do with
    # the call to replace().
    assert repr(t.iso).replace(">U23", "<U23").splitlines() == expected

    # Assign value to unmask, though the instance stays masked.
    t[1] = "2000:111"
    assert str(t) == "['2000:001:00:00:00.000' '2000:111:00:00:00.000']"
    assert t.masked


def test_transform():
    with iers.conf.set_temp("auto_download", False):
        t = Time(["2000:001", "2000:002"])
        t[1] = np.ma.masked

        # Change scale (this tests the ERFA machinery with masking as well)
        t_ut1 = t.ut1
        assert is_masked(t_ut1.value[1])
        assert not is_masked(t_ut1.value[0])
        assert np.all(t_ut1.mask == [False, True])
        # Check the mask is a copy, so we won't back-propagate changes.
        assert not np.may_share_memory(t_ut1.mask, t.mask)

        # Change format
        t_unix = t.unix
        assert is_masked(t_unix[1])
        assert not is_masked(t_unix[0])
        assert np.all(t_unix.mask == [False, True])
        # Check the mask is a copy.
        assert not np.may_share_memory(t_unix.mask, t.mask)


def test_masked_input():
    v0 = np.ma.MaskedArray([[1, 2], [3, 4]])  # No masked elements
    v1 = np.ma.MaskedArray([[1, 2], [3, 4]], mask=[[True, False], [False, False]])
    v2 = np.ma.MaskedArray([[10, 20], [30, 40]], mask=[[False, False], [False, True]])

    # Init from various combinations of masked arrays
    t = Time(v0, format="cxcsec")
    assert np.ma.allclose(t.value, v0)
    assert np.all(t.mask == [[False, False], [False, False]])
    assert t.masked is False

    t = Time(v1, format="cxcsec")
    assert np.ma.allclose(t.value, v1)
    assert np.all(t.mask == v1.mask)
    value = t.value
    assert np.all(value.mask == v1.mask)
    assert t.masked is True
    # Check that masked are not shared with input or output.
    assert not np.may_share_memory(t.mask, v1.mask)
    assert not np.may_share_memory(value.mask, t.mask)
    # But they should be shared in the private _jd1, _jd2.
    assert np.may_share_memory(t._time.jd1.mask, t._time.jd2.mask)

    t = Time(v1, v2, format="cxcsec")
    assert np.ma.allclose(t.value, v1 + v2)
    assert np.all(t.mask == (v1 + v2).mask)
    assert t.masked is True

    t = Time(v0, v1, format="cxcsec")
    assert np.ma.allclose(t.value, v0 + v1)
    assert np.all(t.mask == (v0 + v1).mask)
    assert t.masked is True

    t = Time(0, v2, format="cxcsec")
    assert np.ma.allclose(t.value, v2)
    assert np.all(t.mask == v2.mask)
    assert t.masked is True

    # Init from a string masked array
    t_iso = t.iso
    t2 = Time(t_iso)
    assert np.all(t2.value == t_iso)
    assert np.all(t2.mask == v2.mask)
    assert t2.masked is True


@pytest.mark.parametrize("masked_cls", [np.ma.MaskedArray, Masked])
@pytest.mark.parametrize("val", [0, np.nan, [0], [np.nan]])
def test_all_masked_input(masked_cls, val):
    """Fix for #9612"""
    # Test with jd=0 and jd=np.nan. Both triggered an exception prior to #9624
    # due to astropy.utils.exceptions.ErfaError.
    val = masked_cls(val, mask=True)
    t = Time(val, format="jd")
    if val.ndim:
        assert str(t.iso).endswith("———]")
    else:
        assert str(t.iso).endswith("———")


@pytest.mark.parametrize("mask_cls", [np.ma.MaskedArray, Masked])
@pytest.mark.parametrize("val", ["", [".."], b"", [b".."]])
def test_all_masked_input_str(mask_cls, val):
    dates = mask_cls(val, mask=True)
    t = Time(dates, format="iso", in_subfmt="date")
    assert np.all(t.mask)
    assert np.all(t.unmasked == "2000-01-01")


def test_some_masked_input_str_no_subfmt():
    """Test for cls.fill_value() being longer than other input strings."""
    dates = Masked(["", "2023:001"], mask=True)
    t = Time(dates, format="yday")
    assert np.all(t.mask)
    assert np.all(t.unmasked == "2000:001:12:00:00.000")


def test_masked_input_and_unmasked_array_location():
    dates = Masked(["", "2023:001"], mask=[True, False])
    locations = EarthLocation([10, 20] * u.deg, [30, 40] * u.deg)
    t = Time(dates, format="yday", location=locations)
    assert t.masked
    assert not isinstance(t.location, Masked)
    assert np.all(t.unmasked.location == locations)


def test_serialize_fits_masked(tmp_path):
    tm = Time([1, 2, 3], format="cxcsec")
    tm[1] = np.ma.masked

    fn = tmp_path / "tempfile.fits"
    t = Table([tm])
    t.write(fn)

    t2 = Table.read(fn, astropy_native=True)

    # Time FITS handling does not current round-trip format in FITS
    t2["col0"].format = tm.format

    assert t2["col0"].masked
    assert np.all(t2["col0"].mask == [False, True, False])
    assert np.all(t2["col0"].value == t["col0"].value)


@pytest.mark.skipif(not HAS_H5PY, reason="Needs h5py")
def test_serialize_hdf5_masked(tmp_path):
    tm = Time([1, 2, 3], format="cxcsec")
    tm[1] = np.ma.masked

    fn = tmp_path / "tempfile.hdf5"
    t = Table([tm])
    t.write(fn, path="root", serialize_meta=True)
    t2 = Table.read(fn)

    assert t2["col0"].masked
    assert np.all(t2["col0"].mask == [False, True, False])
    assert np.all(t2["col0"].value == t["col0"].value)


# Ignore warning in MIPS https://github.com/astropy/astropy/issues/9750
@pytest.mark.filterwarnings("ignore:invalid value encountered")
@pytest.mark.parametrize("serialize_method", ["jd1_jd2", "formatted_value"])
def test_serialize_ecsv_masked(serialize_method, tmp_path):
    tm = Time([1, 2, 3], format="cxcsec")
    tm[1] = np.ma.masked

    tm.info.serialize_method["ecsv"] = serialize_method

    fn = tmp_path / "tempfile.ecsv"
    t = Table([tm])
    t.write(fn)
    t2 = Table.read(fn)

    assert t2["col0"].masked
    assert np.all(t2["col0"].mask == [False, True, False])
    # Serializing formatted_value loses some precision.
    atol = 0.1 * u.us if serialize_method == "formatted_value" else 1 * u.ps
    assert np.all(abs(t2["col0"] - t["col0"]) <= atol)


@pytest.mark.parametrize("format_", Time.FORMATS)
@pytest.mark.parametrize("masked_cls", [np.ma.MaskedArray, Masked])
@pytest.mark.parametrize("masked_array_type", ["numpy", "astropy"])
def test_all_formats(format_, masked_cls, masked_array_type):
    mjd = np.array([55000.25, 55000.375, 55001.125])
    mask = np.array([True, False, False])
    mjdm = masked_cls(mjd, mask=mask)
    assert np.may_share_memory(mjdm.mask, mask)
    t = Time(mjd, format="mjd")
    tm = Time(mjdm, format="mjd")
    assert tm.masked and np.all(tm.mask == mask)
    assert not np.may_share_memory(tm.mask, mask)

    out_cls = np.ma.MaskedArray if masked_array_type == "numpy" else Masked
    with conf.set_temp("masked_array_type", masked_array_type):
        # Get values in the given format, check that these have the appropriate class
        # and are correct (ignoring masked ones, which get adjusted on Time
        # initialization, in core._check_for_masked_and_fill).
        t_format = getattr(t, format_)
        tm_format = getattr(tm, format_)
        assert isinstance(tm_format, out_cls)
        if NUMPY_LT_1_26 and format_ == "ymdhms" and out_cls is np.ma.MaskedArray:
            # Work around https://github.com/numpy/numpy/issues/24554
            expected = out_cls(t_format, mask=mask)
        else:
            expected = t_format
        assert np.all(tm_format == expected)

        # Check masked scalar.
        tm0_format = getattr(tm[0], format_)
        assert isinstance(tm0_format, out_cls)
        if NUMPY_LT_1_25 and tm0_format.dtype.kind == "M":
            # Comparison bug in older numpy, just skip it.
            return
        comparison = tm0_format == tm_format[0]
        assert comparison.mask
        if out_cls is Masked:
            assert comparison.unmasked
        elif tm0_format.dtype.names:
            assert comparison.data
        else:
            assert comparison is np.ma.masked

        # Check unmasked scalar.
        tm1_format = getattr(tm[1], format_)
        assert tm1_format == tm_format[1]

    # Verify that configuration gets reset to "astropy".
    tm_format2 = getattr(tm, format_)
    assert isinstance(tm_format2, Masked)

    # Verify that we can also initialize with the format and that this gives
    # the right result and mask too.
    t2 = Time(t_format, format=format_)
    tm2 = Time(tm_format, format=format_)
    assert tm2.masked and np.all(tm2.mask == mask)
    assert np.all(tm2 == t2)


def test_datetime64_with_nat():
    dt64 = np.array(["nat", "2001-02-03", "2001-02-04"], dtype="datetime64")
    mdt64 = Masked(dt64, mask=[False, True, False])
    t = Time(mdt64)
    assert t.masked
    assert np.all(t.mask == [True, True, False])



# Licensed under a 3-clause BSD style license - see LICENSE.rst

import numpy as np
import pytest

from astropy import units as u
from astropy.coordinates import builtin_frames as bf
from astropy.coordinates import galactocentric_frame_defaults
from astropy.coordinates import representation as r
from astropy.coordinates.builtin_frames import CIRS, ICRS, Galactic, Galactocentric
from astropy.coordinates.errors import ConvertError
from astropy.units import allclose as quantity_allclose

POSITION_ON_SKY = {"ra": 37.4 * u.deg, "dec": -55.8 * u.deg}
DISTANCE = {"distance": 150 * u.pc}
PROPER_MOTION = {"pm_ra_cosdec": -21.2 * u.mas / u.yr, "pm_dec": 17.1 * u.mas / u.yr}
RADIAL_VELOCITY = {"radial_velocity": 105.7 * u.km / u.s}

CARTESIAN_POSITION = {
    "x": 1 * u.pc,
    "y": 2 * u.pc,
    "z": 3 * u.pc,
    "representation_type": r.CartesianRepresentation,
}
CARTESIAN_REPRESENTATION_KEYWORD_STR = {"representation_type": "cartesian"}
CARTESIAN_VELOCITY = {
    "v_x": 1 * u.km / u.s,
    "v_y": 2 * u.km / u.s,
    "v_z": 3 * u.km / u.s,
    "differential_type": r.CartesianDifferential,
}
CARTESIAN_DIFFERENTIAL_KEYWORD_STR = {"differential_type": "cartesian"}


def test_api():
    # transform observed Barycentric velocities to full-space Galactocentric
    with galactocentric_frame_defaults.set("latest"):
        icrs = ICRS(**POSITION_ON_SKY, **DISTANCE, **PROPER_MOTION, **RADIAL_VELOCITY)
        icrs.transform_to(Galactocentric())

        # transform a set of ICRS proper motions to Galactic
        ICRS(**POSITION_ON_SKY, **PROPER_MOTION).transform_to(Galactic())


@pytest.mark.parametrize(
    "kwargs",
    [
        POSITION_ON_SKY,
        POSITION_ON_SKY | DISTANCE,
        POSITION_ON_SKY | PROPER_MOTION,
        POSITION_ON_SKY | DISTANCE | PROPER_MOTION,
        POSITION_ON_SKY | RADIAL_VELOCITY,
        POSITION_ON_SKY | DISTANCE | RADIAL_VELOCITY,
        POSITION_ON_SKY | PROPER_MOTION | RADIAL_VELOCITY,
        POSITION_ON_SKY | DISTANCE | PROPER_MOTION | RADIAL_VELOCITY,
        # Now test other representation/differential types:
        CARTESIAN_POSITION,
        CARTESIAN_POSITION | CARTESIAN_REPRESENTATION_KEYWORD_STR,
        CARTESIAN_POSITION | CARTESIAN_VELOCITY,
        CARTESIAN_POSITION | CARTESIAN_VELOCITY | CARTESIAN_DIFFERENTIAL_KEYWORD_STR,
    ],
)
def test_all_arg_options(kwargs):
    # Here we do a simple thing and just verify that passing kwargs in, we have
    # access to the relevant attributes from the resulting object
    icrs = ICRS(**kwargs)
    gal = icrs.transform_to(Galactic())
    repr_gal = repr(gal)

    for k in kwargs:
        if k == "differential_type":
            continue
        getattr(icrs, k)

    if "pm_ra_cosdec" in kwargs:  # should have both
        assert "pm_l_cosb" in repr_gal
        assert "pm_b" in repr_gal
        assert "mas / yr" in repr_gal

        if "radial_velocity" not in kwargs:
            assert "radial_velocity" not in repr_gal

    if "radial_velocity" in kwargs:
        assert "radial_velocity" in repr_gal
        assert "km / s" in repr_gal

        if "pm_ra_cosdec" not in kwargs:
            assert "pm_l_cosb" not in repr_gal
            assert "pm_b" not in repr_gal


@pytest.mark.parametrize(
    "cls,lon,lat",
    [
        [bf.ICRS, "ra", "dec"],
        [bf.FK4, "ra", "dec"],
        [bf.FK4NoETerms, "ra", "dec"],
        [bf.FK5, "ra", "dec"],
        [bf.GCRS, "ra", "dec"],
        [bf.HCRS, "ra", "dec"],
        [bf.LSR, "ra", "dec"],
        [bf.CIRS, "ra", "dec"],
        [bf.Galactic, "l", "b"],
        [bf.AltAz, "az", "alt"],
        [bf.Supergalactic, "sgl", "sgb"],
        [bf.GalacticLSR, "l", "b"],
        [bf.HeliocentricMeanEcliptic, "lon", "lat"],
        [bf.GeocentricMeanEcliptic, "lon", "lat"],
        [bf.BarycentricMeanEcliptic, "lon", "lat"],
        [bf.PrecessedGeocentric, "ra", "dec"],
    ],
)
def test_expected_arg_names(cls, lon, lat):
    kwargs = {
        lon: 37.4 * u.deg,
        lat: -55.8 * u.deg,
        f"pm_{lon}_cos{lat}": -21.2 * u.mas / u.yr,
        f"pm_{lat}": 17.1 * u.mas / u.yr,
    }
    frame = cls(**kwargs, **DISTANCE, **RADIAL_VELOCITY)


# these data are extracted from the vizier version of XHIP:
# https://vizier.cds.unistra.fr/viz-bin/VizieR-3?-source=+V/137A/XHIP
_xhip_head = """
------ ------------ ------------ -------- -------- ------------ ------------ ------- -------- -------- ------- ------ ------ ------
       R            D            pmRA     pmDE                               Di      pmGLon   pmGLat   RV      U      V      W
HIP    AJ2000 (deg) EJ2000 (deg) (mas/yr) (mas/yr) GLon (deg)   GLat (deg)   st (pc) (mas/yr) (mas/yr) (km/s)  (km/s) (km/s) (km/s)
------ ------------ ------------ -------- -------- ------------ ------------ ------- -------- -------- ------- ------ ------ ------
""".strip()

_xhip_data = """
    19 000.05331690 +38.30408633    -3.17   -15.37 112.00026470 -23.47789171  247.12    -6.40   -14.33    6.30    7.3    2.0  -17.9
    20 000.06295067 +23.52928427    36.11   -22.48 108.02779304 -37.85659811   95.90    29.35   -30.78   37.80  -19.3   16.1  -34.2
    21 000.06623581 +08.00723430    61.48    -0.23 101.69697120 -52.74179515  183.68    58.06   -20.23  -11.72  -45.2  -30.9   -1.3
 24917 080.09698238 -33.39874984    -4.30    13.40 236.92324669 -32.58047131  107.38   -14.03    -1.15   36.10  -22.4  -21.3  -19.9
 59207 182.13915108 +65.34963517    18.17     5.49 130.04157185  51.18258601   56.00   -18.98    -0.49    5.70    1.5    6.1    4.4
 87992 269.60730667 +36.87462906   -89.58    72.46  62.98053142  25.90148234  129.60    45.64   105.79   -4.00  -39.5  -15.8   56.7
115110 349.72322473 -28.74087144    48.86    -9.25  23.00447250 -69.52799804  116.87    -8.37   -49.02   15.00  -16.8  -12.2  -23.6
""".strip()

# in principal we could parse the above as a table, but doing it "manually"
# makes this test less tied to Table working correctly


@pytest.mark.parametrize(
    "hip,ra,dec,pmra,pmdec,glon,glat,dist,pmglon,pmglat,rv,U,V,W",
    [[float(val) for val in row.split()] for row in _xhip_data.split("\n")],
)
def test_xhip_galactic(
    hip, ra, dec, pmra, pmdec, glon, glat, dist, pmglon, pmglat, rv, U, V, W
):
    i = ICRS(
        ra * u.deg,
        dec * u.deg,
        dist * u.pc,
        pm_ra_cosdec=pmra * u.marcsec / u.yr,
        pm_dec=pmdec * u.marcsec / u.yr,
        radial_velocity=rv * u.km / u.s,
    )
    g = i.transform_to(Galactic())

    # precision is limited by 2-deciimal digit string representation of pms
    assert quantity_allclose(
        g.pm_l_cosb, pmglon * u.marcsec / u.yr, atol=0.01 * u.marcsec / u.yr
    )
    assert quantity_allclose(
        g.pm_b, pmglat * u.marcsec / u.yr, atol=0.01 * u.marcsec / u.yr
    )

    # make sure UVW also makes sense
    uvwg = g.cartesian.differentials["s"]
    # precision is limited by 1-decimal digit string representation of vels
    assert quantity_allclose(uvwg.d_x, U * u.km / u.s, atol=0.1 * u.km / u.s)
    assert quantity_allclose(uvwg.d_y, V * u.km / u.s, atol=0.1 * u.km / u.s)
    assert quantity_allclose(uvwg.d_z, W * u.km / u.s, atol=0.1 * u.km / u.s)


@pytest.mark.parametrize(
    "kwargs,expect_success",
    (
        (POSITION_ON_SKY, False),
        (POSITION_ON_SKY | DISTANCE, True),
        (POSITION_ON_SKY | PROPER_MOTION, False),
        (POSITION_ON_SKY | RADIAL_VELOCITY, False),
        (POSITION_ON_SKY | DISTANCE | RADIAL_VELOCITY, False),
        (POSITION_ON_SKY | PROPER_MOTION | RADIAL_VELOCITY, False),
        (POSITION_ON_SKY | DISTANCE | PROPER_MOTION | RADIAL_VELOCITY, True),
    ),
)
def test_frame_affinetransform(kwargs, expect_success):
    """There are already tests in test_transformations.py that check that
    an AffineTransform fails without full-space data, but this just checks that
    things work as expected at the frame level as well.
    """
    with galactocentric_frame_defaults.set("latest"):
        icrs = ICRS(**kwargs)

        if expect_success:
            _ = icrs.transform_to(Galactocentric())

        else:
            with pytest.raises(ConvertError):
                icrs.transform_to(Galactocentric())


def test_differential_type_arg():
    """
    Test passing in an explicit differential class to the initializer or
    changing the differential class via set_representation_cls
    """
    icrs = ICRS(
        **POSITION_ON_SKY,
        pm_ra=10 * u.mas / u.yr,
        pm_dec=-11 * u.mas / u.yr,
        differential_type=r.UnitSphericalDifferential,
    )
    assert icrs.pm_ra == 10 * u.mas / u.yr

    icrs = ICRS(
        **POSITION_ON_SKY,
        pm_ra=10 * u.mas / u.yr,
        pm_dec=-11 * u.mas / u.yr,
        differential_type={"s": r.UnitSphericalDifferential},
    )
    assert icrs.pm_ra == 10 * u.mas / u.yr

    icrs = ICRS(
        ra=1 * u.deg,
        dec=60 * u.deg,
        pm_ra_cosdec=10 * u.mas / u.yr,
        pm_dec=-11 * u.mas / u.yr,
    )
    icrs.set_representation_cls(s=r.UnitSphericalDifferential)
    assert quantity_allclose(icrs.pm_ra, 20 * u.mas / u.yr)

    # incompatible representation and differential
    with pytest.raises(TypeError):
        ICRS(**POSITION_ON_SKY, **CARTESIAN_VELOCITY)

    # specify both
    icrs = ICRS(**CARTESIAN_POSITION, **CARTESIAN_VELOCITY)
    assert icrs.x == 1 * u.pc
    assert icrs.y == 2 * u.pc
    assert icrs.z == 3 * u.pc
    assert icrs.v_x == 1 * u.km / u.s
    assert icrs.v_y == 2 * u.km / u.s
    assert icrs.v_z == 3 * u.km / u.s


def test_slicing_preserves_differential():
    icrs = ICRS(**POSITION_ON_SKY, **DISTANCE, **PROPER_MOTION, **RADIAL_VELOCITY)
    icrs2 = icrs.reshape(1, 1)[:1, 0]

    for name in icrs.representation_component_names.keys():
        assert getattr(icrs, name) == getattr(icrs2, name)[0]

    for name in icrs.get_representation_component_names("s").keys():
        assert getattr(icrs, name) == getattr(icrs2, name)[0]


def test_shorthand_attributes():
    # Check that attribute access works

    # for array data:
    n = 4
    icrs1 = ICRS(
        ra=np.random.uniform(0, 360, n) * u.deg,
        dec=np.random.uniform(-90, 90, n) * u.deg,
        distance=100 * u.pc,
        pm_ra_cosdec=np.random.normal(0, 100, n) * u.mas / u.yr,
        pm_dec=np.random.normal(0, 100, n) * u.mas / u.yr,
        radial_velocity=np.random.normal(0, 100, n) * u.km / u.s,
    )
    v = icrs1.velocity
    pm = icrs1.proper_motion
    assert quantity_allclose(pm[0], icrs1.pm_ra_cosdec)
    assert quantity_allclose(pm[1], icrs1.pm_dec)

    # for scalar data:
    icrs2 = ICRS(**POSITION_ON_SKY, **DISTANCE, **PROPER_MOTION, **RADIAL_VELOCITY)
    v = icrs2.velocity
    pm = icrs2.proper_motion
    assert quantity_allclose(pm[0], icrs2.pm_ra_cosdec)
    assert quantity_allclose(pm[1], icrs2.pm_dec)

    # check that it fails where we expect:

    # no distance
    icrs3 = ICRS(**POSITION_ON_SKY, **PROPER_MOTION, **RADIAL_VELOCITY)
    with pytest.raises(ValueError):
        icrs3.velocity

    icrs3.set_representation_cls("cartesian")
    assert hasattr(icrs3, "radial_velocity")
    assert quantity_allclose(icrs3.radial_velocity, 105.7 * u.km / u.s)

    icrs4 = ICRS(**CARTESIAN_POSITION, **CARTESIAN_VELOCITY)
    icrs4.radial_velocity


@pytest.mark.parametrize(
    "icrs_coords", [POSITION_ON_SKY, POSITION_ON_SKY | PROPER_MOTION]
)
def test_negative_distance(icrs_coords):
    """Regression test: #7408
    Make sure that negative parallaxes turned into distances are handled right
    """
    c = ICRS(distance=(-10 * u.mas).to(u.pc, u.parallax()), **icrs_coords)
    assert quantity_allclose(c.ra, 37.4 * u.deg)
    assert quantity_allclose(c.dec, -55.8 * u.deg)


def test_velocity_units():
    """Check that the differential data given has compatible units
    with the time-derivative of representation data"""
    with pytest.raises(
        ValueError,
        match=(
            '^x has unit "pc" with physical type "length", but v_x has incompatible'
            ' unit "" with physical type "dimensionless" instead of the expected'
            r' "speed/velocity".$'
        ),
    ):
        ICRS(**CARTESIAN_POSITION, v_x=1, v_y=2, v_z=3, differential_type="cartesian")


def test_frame_with_velocity_without_distance_can_be_transformed():
    rep = CIRS(**POSITION_ON_SKY, **PROPER_MOTION).transform_to(ICRS())
    assert "<ICRS Coordinate: (ra, dec, distance) in" in repr(rep)



# Licensed under a 3-clause BSD style license - see LICENSE.rst

__all__: list[str] = ["ParametersAttribute"]

from dataclasses import dataclass, field
from types import MappingProxyType
from typing import TYPE_CHECKING, Any, NoReturn, Union

if TYPE_CHECKING:
    import astropy.cosmology


@dataclass(frozen=True, slots=True)
class ParametersAttribute:
    """Immutable mapping of the :class:`~astropy.cosmology.Parameter` objects or values.

    If accessed from the :class:`~astropy.cosmology.Cosmology` class, this returns a
    mapping of the :class:`~astropy.cosmology.Parameter` objects themselves.  If
    accessed from an instance, this returns a mapping of the values of the Parameters.

    This class is used to implement :obj:`astropy.cosmology.Cosmology.parameters`.

    Parameters
    ----------
    attr_name : str
        The name of the class attribute that is a `~types.MappingProxyType[str,
        astropy.cosmology.Parameter]` of all the cosmology's parameters. When accessed
        from the class, this attribute is returned. When accessed from an instance, a
        mapping of the cosmology instance's values for each key is returned.

    Examples
    --------
    The normal usage of this class is the ``parameters`` attribute of
    :class:`~astropy.cosmology.Cosmology`.

        >>> from astropy.cosmology import FlatLambdaCDM, Planck18

        >>> FlatLambdaCDM.parameters
        mappingproxy({'H0': Parameter(...), ...})

        >>> Planck18.parameters
        mappingproxy({'H0': <Quantity 67.66 km / (Mpc s)>, ...})
    """

    attr_name: str
    """Class attribute name on Cosmology for the mapping of Parameter objects."""

    _name: str = field(init=False)
    """The name of the descriptor on the containing class."""

    def __set_name__(self, owner: Any, name: str) -> None:
        object.__setattr__(self, "_name", name)

    def __get__(
        self,
        instance: Union["astropy.cosmology.Cosmology", None],
        owner: type["astropy.cosmology.Cosmology"] | None,
    ) -> MappingProxyType[str, Any]:
        # Called from the class
        if instance is None:
            return getattr(owner, self.attr_name)
        # Called from the instance
        return MappingProxyType(
            {n: getattr(instance, n) for n in getattr(instance, self.attr_name)}
        )

    def __set__(self, instance: Any, value: Any) -> NoReturn:
        msg = f"cannot set {self._name!r} of {instance!r}."
        raise AttributeError(msg)



# Licensed under a 3-clause BSD style license - see LICENSE.rst
"""
Mixin columns for use in ascii/tests/test_ecsv.py, fits/tests/test_connect.py,
and misc/tests/test_hdf5.py.

All columns should have length 2.
"""

import numpy as np

from astropy import coordinates, table, time
from astropy import units as u

el = coordinates.EarthLocation(x=[1, 2] * u.km, y=[3, 4] * u.km, z=[5, 6] * u.km)
sr = coordinates.SphericalRepresentation([0, 1] * u.deg, [2, 3] * u.deg, 1 * u.kpc)
cr = coordinates.CartesianRepresentation([0, 1] * u.pc, [4, 5] * u.pc, [8, 6] * u.pc)
sd = coordinates.SphericalCosLatDifferential(
    [0, 1] * u.mas / u.yr, [0, 1] * u.mas / u.yr, 10 * u.km / u.s
)
srd = coordinates.SphericalRepresentation(sr, differentials=sd)
sc = coordinates.SkyCoord(
    [1, 2], [3, 4], unit="deg,deg", frame="fk4", obstime="J1990.5"
)
scd = coordinates.SkyCoord(
    [1, 2], [3, 4], [5, 6], unit="deg,deg,m", frame="fk4", obstime=["J1990.5"] * 2
)
scdc = scd.copy()
scdc.representation_type = "cartesian"
scpm = coordinates.SkyCoord(
    [1, 2],
    [3, 4],
    [5, 6],
    unit="deg,deg,pc",
    pm_ra_cosdec=[7, 8] * u.mas / u.yr,
    pm_dec=[9, 10] * u.mas / u.yr,
)
scpmrv = coordinates.SkyCoord(
    [1, 2],
    [3, 4],
    [5, 6],
    unit="deg,deg,pc",
    pm_ra_cosdec=[7, 8] * u.mas / u.yr,
    pm_dec=[9, 10] * u.mas / u.yr,
    radial_velocity=[11, 12] * u.km / u.s,
)
scrv = coordinates.SkyCoord(
    [1, 2], [3, 4], [5, 6], unit="deg,deg,pc", radial_velocity=[11, 12] * u.km / u.s
)
# Use UnitSpherical and Radial for a regression test for gh-16998.
icrs = coordinates.ICRS(
    [1, 2] * u.deg,
    [3, 4] * u.deg,
    radial_velocity=[5, 6] * u.km / u.s,
    representation_type="unitspherical",
    differential_type="radial",
)
altaz = coordinates.AltAz(
    [1, 2] * u.deg,
    [3, 4] * u.deg,
    obstime=time.Time("2010-11-12T13:14:15"),
    location=el,
)
so = coordinates.SkyOffsetFrame(
    [-1, -2] * u.deg,
    [-3, -4] * u.deg,
    origin=altaz,
    rotation=[5, 6] * u.deg,
)
sond = coordinates.SkyOffsetFrame(
    origin=icrs,
    rotation=[5, 6] * u.deg,
)

tm = time.Time(
    [51000.5, 51001.5], format="mjd", scale="tai", precision=5, location=el[0]
)
tm2 = time.Time(tm, precision=3, format="iso")
tm3 = time.Time(tm, location=el)
tm3.info.serialize_method["ecsv"] = "jd1_jd2"
obj = table.Column([{"a": 1}, {"b": [2]}], dtype="object")
su = table.Column(
    [(1, (1.5, 1.6)), (2, (2.5, 2.6))],
    name="su",
    dtype=[("i", np.int64), ("f", [("p1", np.float64), ("p0", np.float64)])],
)
su2 = table.Column(
    [(["snake", "c"], [1.6, 1.5]), (["eal", "a"], [2.5, 2.6])],
    dtype=[("name", "U5", (2,)), ("f", "f8", (2,))],
)
stokes = coordinates.StokesCoord(["RR", "LL"])

# NOTE: for testing, the name of the column "x" for the
# Quantity is important since it tests the fix for #10215
# (namespace clash, where "x" clashes with "el.x").
mixin_cols = {
    "tm": tm,
    "tm2": tm2,
    "tm3": tm3,
    "dt": time.TimeDelta([1, 2] * u.day),
    "sc": sc,
    "scd": scd,
    "scdc": scdc,
    "scpm": scpm,
    "scpmrv": scpmrv,
    "scrv": scrv,
    "icrs": icrs,
    "altaz": altaz,
    "so": so,
    "sond": sond,
    "x": [1, 2] * u.m,
    "qdb": [10, 20] * u.dB(u.mW),
    "qdex": [4.5, 5.5] * u.dex(u.cm / u.s**2),
    "qmag": [21, 22] * u.ABmag,
    "lat": coordinates.Latitude([1, 2] * u.deg),
    "lon": coordinates.Longitude([1, 2] * u.deg, wrap_angle=180.0 * u.deg),
    "ang": coordinates.Angle([1, 2] * u.deg),
    "el": el,
    "sr": sr,
    "cr": cr,
    "sd": sd,
    "srd": srd,
    "nd": table.NdarrayMixin([1, 2]),
    "obj": obj,
    "su": su,
    "su2": su2,
    "stokes": stokes,
}
time_attrs = [
    "value",
    "shape",
    "format",
    "scale",
    "precision",
    "in_subfmt",
    "out_subfmt",
    "location",
]
compare_attrs = {
    "tm": time_attrs,
    "tm2": time_attrs,
    "tm3": time_attrs,
    "dt": ["shape", "value", "format", "scale"],
    "sc": ["ra", "dec", "representation_type", "frame.name"],
    "scd": ["ra", "dec", "distance", "representation_type", "frame.name"],
    "scdc": ["x", "y", "z", "representation_type", "frame.name"],
    "scpm": [
        "ra",
        "dec",
        "distance",
        "pm_ra_cosdec",
        "pm_dec",
        "representation_type",
        "frame.name",
    ],
    "scpmrv": [
        "ra",
        "dec",
        "distance",
        "pm_ra_cosdec",
        "pm_dec",
        "radial_velocity",
        "representation_type",
        "frame.name",
    ],
    "scrv": [
        "ra",
        "dec",
        "distance",
        "radial_velocity",
        "representation_type",
        "frame.name",
    ],
    "icrs": [
        "ra",
        "dec",
        "radial_velocity",
        "representation_type",
        "differential_type",
    ],
    "altaz": ["alt", "az", "obstime", "location", "representation_type"],
    "so": ["lon", "lat", "rotation", "origin"],
    "sond": ["rotation", "origin"],
    "x": ["value", "unit"],
    "qdb": ["value", "unit"],
    "qdex": ["value", "unit"],
    "qmag": ["value", "unit"],
    "lon": ["value", "unit", "wrap_angle"],
    "lat": ["value", "unit"],
    "ang": ["value", "unit"],
    "el": ["x", "y", "z", "ellipsoid"],
    "nd": ["data"],
    "sr": ["lon", "lat", "distance"],
    "cr": ["x", "y", "z"],
    "sd": ["d_lon_coslat", "d_lat", "d_distance"],
    "srd": [
        "lon",
        "lat",
        "distance",
        "differentials.s.d_lon_coslat",
        "differentials.s.d_lat",
        "differentials.s.d_distance",
    ],
    "obj": [],
    "su": ["i", "f.p0", "f.p1"],
    "su2": ["name", "f"],
    "stokes": ["value"],
}
non_trivial_names = {
    "cr": ["cr.x", "cr.y", "cr.z"],
    "dt": ["dt.jd1", "dt.jd2"],
    "el": ["el.x", "el.y", "el.z"],
    "sc": ["sc.ra", "sc.dec"],
    "scd": ["scd.ra", "scd.dec", "scd.distance", "scd.obstime.jd1", "scd.obstime.jd2"],
    "scdc": ["scdc.x", "scdc.y", "scdc.z", "scdc.obstime.jd1", "scdc.obstime.jd2"],
    "scfc": ["scdc.x", "scdc.y", "scdc.z", "scdc.obstime.jd1", "scdc.obstime.jd2"],
    "scpm": [
        "scpm.ra",
        "scpm.dec",
        "scpm.distance",
        "scpm.pm_ra_cosdec",
        "scpm.pm_dec",
    ],
    "scpmrv": [
        "scpmrv.ra",
        "scpmrv.dec",
        "scpmrv.distance",
        "scpmrv.pm_ra_cosdec",
        "scpmrv.pm_dec",
        "scpmrv.radial_velocity",
    ],
    "scrv": ["scrv.ra", "scrv.dec", "scrv.distance", "scrv.radial_velocity"],
    "icrs": ["icrs.ra", "icrs.dec", "icrs.radial_velocity"],
    "altaz": [
        "altaz.az",
        "altaz.alt",
        "altaz.location.x",
        "altaz.location.y",
        "altaz.location.z",
    ],
    "so": [
        "so.lon",
        "so.lat",
        "so.rotation",
        "so.origin.az",
        "so.origin.alt",
        "so.origin.location.x",
        "so.origin.location.y",
        "so.origin.location.z",
    ],
    "sond": [
        "sond.rotation",
        "sond.origin.ra",
        "sond.origin.dec",
        "sond.origin.radial_velocity",
    ],
    "sd": ["sd.d_lon_coslat", "sd.d_lat", "sd.d_distance"],
    "sr": ["sr.lon", "sr.lat", "sr.distance"],
    "srd": [
        "srd.lon",
        "srd.lat",
        "srd.distance",
        "srd.differentials.s.d_lon_coslat",
        "srd.differentials.s.d_lat",
        "srd.differentials.s.d_distance",
    ],
    "su": ["su.i", "su.f.p1", "su.f.p0"],
    "su2": ["su2.name", "su2.f"],
    "tm": ["tm.jd1", "tm.jd2"],
    "tm2": ["tm2.jd1", "tm2.jd2"],
    "tm3": ["tm3.jd1", "tm3.jd2", "tm3.location.x", "tm3.location.y", "tm3.location.z"],
}
serialized_names = {
    name: non_trivial_names.get(name, [name]) for name in sorted(mixin_cols)
}



# Licensed under a 3-clause BSD style license - see LICENSE.rst

"""
This module implements a class based on NDData with all Mixins.
"""

from .mixins.ndarithmetic import NDArithmeticMixin
from .mixins.ndio import NDIOMixin
from .mixins.ndslicing import NDSlicingMixin
from .nddata import NDData

__all__ = ["NDDataRef"]


class NDDataRef(NDArithmeticMixin, NDIOMixin, NDSlicingMixin, NDData):
    """Implements `NDData` with all Mixins.

    This class implements a `NDData`-like container that supports reading and
    writing as implemented in the ``astropy.io.registry`` and also slicing
    (indexing) and simple arithmetic (add, subtract, divide and multiply).

    Notes
    -----
    A key distinction from `NDDataArray` is that this class does not attempt
    to provide anything that was not defined in any of the parent classes.

    See Also
    --------
    NDData
    NDArithmeticMixin
    NDSlicingMixin
    NDIOMixin

    Examples
    --------
    The mixins allow operation that are not possible with `NDData` or
    `NDDataBase`, i.e. simple arithmetic::

        >>> from astropy.nddata import NDDataRef, StdDevUncertainty
        >>> import numpy as np

        >>> data = np.ones((3,3), dtype=float)
        >>> ndd1 = NDDataRef(data, uncertainty=StdDevUncertainty(data))
        >>> ndd2 = NDDataRef(data, uncertainty=StdDevUncertainty(data))

        >>> ndd3 = ndd1.add(ndd2)
        >>> ndd3.data  # doctest: +FLOAT_CMP
        array([[2., 2., 2.],
               [2., 2., 2.],
               [2., 2., 2.]])
        >>> ndd3.uncertainty.array  # doctest: +FLOAT_CMP
        array([[1.41421356, 1.41421356, 1.41421356],
               [1.41421356, 1.41421356, 1.41421356],
               [1.41421356, 1.41421356, 1.41421356]])

    see `NDArithmeticMixin` for a complete list of all supported arithmetic
    operations.

    But also slicing (indexing) is possible::

        >>> ndd4 = ndd3[1,:]
        >>> ndd4.data  # doctest: +FLOAT_CMP
        array([2., 2., 2.])
        >>> ndd4.uncertainty.array  # doctest: +FLOAT_CMP
        array([1.41421356, 1.41421356, 1.41421356])

    See `NDSlicingMixin` for a description how slicing works (which attributes)
    are sliced.
    """



# Licensed under a 3-clause BSD style license - see LICENSE.rst

import warnings

import numpy as np
import pytest

from astropy import units as u
from astropy.table import Column, QTable, Row, Table, hstack
from astropy.table.bst import BST
from astropy.table.column import BaseColumn
from astropy.table.index import SlicedIndex, get_index
from astropy.table.soco import SCEngine
from astropy.table.sorted_array import SortedArray
from astropy.time import Time
from astropy.utils.compat.optional_deps import HAS_SORTEDCONTAINERS
from astropy.utils.exceptions import AstropyDeprecationWarning

from .test_table import SetupData

available_engines = [BST, SortedArray]

if HAS_SORTEDCONTAINERS:
    available_engines.append(SCEngine)


@pytest.fixture(params=available_engines)
def engine(request):
    return request.param


_col = [1, 2, 3, 4, 5]


@pytest.fixture(
    params=[
        _col,
        u.Quantity(_col),
        Time(_col, format="jyear"),
    ]
)
def main_col(request):
    return request.param


def assert_col_equal(col, array):
    __tracebackhide__ = True
    if isinstance(col, Time):
        assert np.all(col == Time(array, format="jyear"))
    else:
        assert np.all(col == col.__class__(array))


@pytest.mark.usefixtures("table_types")
class TestIndex(SetupData):
    def _setup(self, main_col, table_types):
        super()._setup(table_types)
        self.main_col = main_col
        if isinstance(main_col, u.Quantity):
            self._table_type = QTable
        if not isinstance(main_col, list):
            self._column_type = lambda x: x  # don't change mixin type
        self.mutable = isinstance(main_col, (list, u.Quantity))

    def make_col(self, name, lst):
        return self._column_type(lst, name=name)

    def make_val(self, val):
        if isinstance(self.main_col, Time):
            return Time(val, format="jyear")
        return val

    @property
    def t(self):
        if not hasattr(self, "_t"):
            # Note that order of columns is important, and the 'a' column is
            # last to ensure that the index column does not need to be the first
            # column (as was discovered in #10025).  Most testing uses 'a' and
            # ('a', 'b') for the columns.
            self._t = self._table_type()
            self._t["b"] = self._column_type([4.0, 5.1, 6.2, 7.0, 1.1])
            self._t["c"] = self._column_type(["7", "8", "9", "10", "11"])
            self._t["a"] = self._column_type(self.main_col)
        return self._t

    @pytest.mark.parametrize("composite", [False, True])
    def test_table_index(self, main_col, table_types, composite, engine):
        self._setup(main_col, table_types)
        t = self.t
        t.add_index(("a", "b") if composite else "a", engine=engine)
        assert np.all(t.indices[0].sorted_data() == [0, 1, 2, 3, 4])

        if not self.mutable:
            return

        # test altering table columns
        t["a"][0] = 4
        t.add_row((6.0, "7", 6))
        t["a"][3] = 10
        t.remove_row(2)
        t.add_row((5.0, "9", 4))

        assert_col_equal(t["a"], np.array([4, 2, 10, 5, 6, 4]))
        assert np.allclose(t["b"], np.array([4.0, 5.1, 7.0, 1.1, 6.0, 5.0]))
        assert np.all(t["c"].data == np.array(["7", "8", "10", "11", "7", "9"]))
        index = t.indices[0]
        ll = list(index.data.items())

        if composite:
            assert np.all(
                ll
                == [
                    ((2, 5.1), [1]),
                    ((4, 4.0), [0]),
                    ((4, 5.0), [5]),
                    ((5, 1.1), [3]),
                    ((6, 6.0), [4]),
                    ((10, 7.0), [2]),
                ]
            )
        else:
            assert np.all(
                ll
                == [((2,), [1]), ((4,), [0, 5]), ((5,), [3]), ((6,), [4]), ((10,), [2])]
            )
        t.remove_indices("a")
        assert len(t.indices) == 0

    def test_table_slicing(self, main_col, table_types, engine):
        self._setup(main_col, table_types)
        t = self.t
        t.add_index("a", engine=engine)
        assert np.all(t.indices[0].sorted_data() == [0, 1, 2, 3, 4])

        for slice_ in ([0, 2], np.array([0, 2])):
            t2 = t[slice_]
            # t2 should retain an index on column 'a'
            assert len(t2.indices) == 1
            assert_col_equal(t2["a"], [1, 3])

            # the index in t2 should reorder row numbers after slicing
            assert np.all(t2.indices[0].sorted_data() == [0, 1])
            # however, this index should be a deep copy of t1's index
            assert np.all(t.indices[0].sorted_data() == [0, 1, 2, 3, 4])

    def test_remove_rows(self, main_col, table_types, engine):
        self._setup(main_col, table_types)
        if not self.mutable:
            return
        t = self.t
        t.add_index("a", engine=engine)

        # remove individual row
        t2 = t.copy()
        t2.remove_rows(2)
        assert_col_equal(t2["a"], [1, 2, 4, 5])
        assert np.all(t2.indices[0].sorted_data() == [0, 1, 2, 3])

        # remove by list, ndarray, or slice
        for cut in ([0, 2, 4], np.array([0, 2, 4]), slice(0, 5, 2)):
            t2 = t.copy()
            t2.remove_rows(cut)
            assert_col_equal(t2["a"], [2, 4])
            assert np.all(t2.indices[0].sorted_data() == [0, 1])

        with pytest.raises(ValueError):
            t.remove_rows((0, 2, 4))

    def test_col_get_slice(self, main_col, table_types, engine):
        self._setup(main_col, table_types)
        t = self.t
        t.add_index("a", engine=engine)

        # get slice
        t2 = t[1:3]  # table slice
        assert_col_equal(t2["a"], [2, 3])
        assert np.all(t2.indices[0].sorted_data() == [0, 1])

        col_slice = t["a"][1:3]
        assert_col_equal(col_slice, [2, 3])
        # true column slices discard indices
        if isinstance(t["a"], BaseColumn):
            assert len(col_slice.info.indices) == 0

        # take slice of slice
        t2 = t[::2]
        assert_col_equal(t2["a"], np.array([1, 3, 5]))
        t3 = t2[::-1]
        assert_col_equal(t3["a"], np.array([5, 3, 1]))
        assert np.all(t3.indices[0].sorted_data() == [2, 1, 0])
        t3 = t2[:2]
        assert_col_equal(t3["a"], np.array([1, 3]))
        assert np.all(t3.indices[0].sorted_data() == [0, 1])
        # out-of-bound slices
        for t_empty in (t2[3:], t2[2:1], t3[2:]):
            assert len(t_empty["a"]) == 0
            assert np.all(t_empty.indices[0].sorted_data() == [])

        if self.mutable:
            # get boolean mask
            mask = t["a"] % 2 == 1
            t2 = t[mask]
            assert_col_equal(t2["a"], [1, 3, 5])
            assert np.all(t2.indices[0].sorted_data() == [0, 1, 2])

    def test_col_set_slice(self, main_col, table_types, engine):
        self._setup(main_col, table_types)
        if not self.mutable:
            return
        t = self.t
        t.add_index("a", engine=engine)

        # set slice
        t2 = t.copy()
        t2["a"][1:3] = np.array([6, 7])
        assert_col_equal(t2["a"], np.array([1, 6, 7, 4, 5]))
        assert np.all(t2.indices[0].sorted_data() == [0, 3, 4, 1, 2])

        # change original table via slice reference
        t2 = t.copy()
        t3 = t2[1:3]
        assert_col_equal(t3["a"], np.array([2, 3]))
        assert np.all(t3.indices[0].sorted_data() == [0, 1])
        t3["a"][0] = 5
        assert_col_equal(t3["a"], np.array([5, 3]))
        assert_col_equal(t2["a"], np.array([1, 5, 3, 4, 5]))
        assert np.all(t3.indices[0].sorted_data() == [1, 0])
        assert np.all(t2.indices[0].sorted_data() == [0, 2, 3, 1, 4])

        # set boolean mask
        t2 = t.copy()
        mask = t["a"] % 2 == 1
        t2["a"][mask] = 0.0
        assert_col_equal(t2["a"], [0, 2, 0, 4, 0])
        assert np.all(t2.indices[0].sorted_data() == [0, 2, 4, 1, 3])

    def test_multiple_slices(self, main_col, table_types, engine):
        self._setup(main_col, table_types)

        if not self.mutable:
            return

        t = self.t
        t.add_index("a", engine=engine)

        for i in range(6, 51):
            t.add_row((1.0, "A", i))

        assert_col_equal(t["a"], list(range(1, 51)))
        assert np.all(t.indices[0].sorted_data() == list(range(50)))

        evens = t[::2]
        assert np.all(evens.indices[0].sorted_data() == list(range(25)))
        reverse = evens[::-1]
        index = reverse.indices[0]
        assert (index.start, index.stop, index.step) == (48, -2, -2)
        assert np.all(index.sorted_data() == list(range(24, -1, -1)))

        # modify slice of slice
        reverse[-10:] = 0
        expected = np.array(list(range(1, 51)))
        expected[:20][expected[:20] % 2 == 1] = 0
        assert_col_equal(t["a"], expected)
        assert_col_equal(evens["a"], expected[::2])
        assert_col_equal(reverse["a"], expected[::2][::-1])
        # first ten evens are now zero
        assert np.all(
            t.indices[0].sorted_data()
            == (
                [0, 2, 4, 6, 8, 10, 12, 14, 16, 18, 1, 3, 5, 7, 9, 11, 13, 15, 17, 19]
                + list(range(20, 50))
            )
        )
        assert np.all(evens.indices[0].sorted_data() == list(range(25)))
        assert np.all(reverse.indices[0].sorted_data() == list(range(24, -1, -1)))

        # try different step sizes of slice
        t2 = t[1:20:2]
        assert_col_equal(t2["a"], [2, 4, 6, 8, 10, 12, 14, 16, 18, 20])
        assert np.all(t2.indices[0].sorted_data() == list(range(10)))
        t3 = t2[::3]
        assert_col_equal(t3["a"], [2, 8, 14, 20])
        assert np.all(t3.indices[0].sorted_data() == [0, 1, 2, 3])
        t4 = t3[2::-1]
        assert_col_equal(t4["a"], [14, 8, 2])
        assert np.all(t4.indices[0].sorted_data() == [2, 1, 0])

    def test_sort(self, main_col, table_types, engine):
        self._setup(main_col, table_types)
        t = self.t[::-1]  # reverse table
        assert_col_equal(t["a"], [5, 4, 3, 2, 1])
        t.add_index("a", engine=engine)
        assert np.all(t.indices[0].sorted_data() == [4, 3, 2, 1, 0])

        if not self.mutable:
            return

        # sort table by column a
        t2 = t.copy()
        t2.sort("a")
        assert_col_equal(t2["a"], [1, 2, 3, 4, 5])
        assert np.all(t2.indices[0].sorted_data() == [0, 1, 2, 3, 4])

        # sort table by primary key
        t2 = t.copy()
        t2.sort()
        assert_col_equal(t2["a"], [1, 2, 3, 4, 5])
        assert np.all(t2.indices[0].sorted_data() == [0, 1, 2, 3, 4])

    def test_insert_row(self, main_col, table_types, engine):
        self._setup(main_col, table_types)

        if not self.mutable:
            return

        t = self.t
        t.add_index("a", engine=engine)
        t.insert_row(2, (1.0, "12", 6))
        assert_col_equal(t["a"], [1, 2, 6, 3, 4, 5])
        assert np.all(t.indices[0].sorted_data() == [0, 1, 3, 4, 5, 2])
        t.insert_row(1, (4.0, "13", 0))
        assert_col_equal(t["a"], [1, 0, 2, 6, 3, 4, 5])
        assert np.all(t.indices[0].sorted_data() == [1, 0, 2, 4, 5, 6, 3])

    def test_index_modes(self, main_col, table_types, engine):
        self._setup(main_col, table_types)
        t = self.t
        t.add_index("a", engine=engine)

        # first, no special mode
        assert len(t[[1, 3]].indices) == 1
        assert len(t[::-1].indices) == 1
        assert len(self._table_type(t).indices) == 1
        assert np.all(t.indices[0].sorted_data() == [0, 1, 2, 3, 4])
        t2 = t.copy()

        # non-copy mode
        with t.index_mode("discard_on_copy"):
            assert len(t[[1, 3]].indices) == 0
            assert len(t[::-1].indices) == 0
            assert len(self._table_type(t).indices) == 0
            assert len(t2.copy().indices) == 1  # mode should only affect t

        # make sure non-copy mode is exited correctly
        assert len(t[[1, 3]].indices) == 1

        if not self.mutable:
            return

        # non-modify mode
        with t.index_mode("freeze"):
            assert np.all(t.indices[0].sorted_data() == [0, 1, 2, 3, 4])
            t["a"][0] = 6
            assert np.all(t.indices[0].sorted_data() == [0, 1, 2, 3, 4])
            t.add_row((1.5, "12", 2))
            assert np.all(t.indices[0].sorted_data() == [0, 1, 2, 3, 4])
            t.remove_rows([1, 3])
            assert np.all(t.indices[0].sorted_data() == [0, 1, 2, 3, 4])
            assert_col_equal(t["a"], [6, 3, 5, 2])
            # mode should only affect t
            assert np.all(t2.indices[0].sorted_data() == [0, 1, 2, 3, 4])
            t2["a"][0] = 6
            assert np.all(t2.indices[0].sorted_data() == [1, 2, 3, 4, 0])

        # make sure non-modify mode is exited correctly
        assert np.all(t.indices[0].sorted_data() == [3, 1, 2, 0])

        if isinstance(t["a"], BaseColumn):
            assert len(t["a"][::-1].info.indices) == 0
            with t.index_mode("copy_on_getitem"):
                assert len(t["a"][[1, 2]].info.indices) == 1
                # mode should only affect t
                assert len(t2["a"][[1, 2]].info.indices) == 0

            assert len(t["a"][::-1].info.indices) == 0
            assert len(t2["a"][::-1].info.indices) == 0

    def test_index_retrieval(self, main_col, table_types, engine):
        self._setup(main_col, table_types)
        t = self.t
        t.add_index("a", engine=engine)
        t.add_index(["a", "c"], engine=engine)
        assert len(t.indices) == 2
        assert len(t.indices["a"].columns) == 1
        assert len(t.indices["a", "c"].columns) == 2

        with pytest.raises(IndexError):
            t.indices["b"]

    def test_col_rename(self, main_col, table_types, engine):
        """
        Checks for a previous bug in which copying a Table
        with different column names raised an exception.
        """
        self._setup(main_col, table_types)
        t = self.t
        t.add_index("a", engine=engine)
        t2 = self._table_type(self.t, names=["d", "e", "f"])
        assert len(t2.indices) == 1

    def test_table_loc(self, main_col, table_types, engine):
        self._setup(main_col, table_types)
        t = self.t

        t.add_index("a", engine=engine)
        t.add_index("b", engine=engine)

        t2 = t.loc[self.make_val(3)]  # single label, with primary key 'a'
        assert_col_equal(t2["a"], [3])
        assert isinstance(t2, Row)

        # list search
        t2 = t.loc[[self.make_val(1), self.make_val(4), self.make_val(2)]]
        assert_col_equal(t2["a"], [1, 4, 2])  # same order as input list
        if not isinstance(main_col, Time):
            # ndarray search
            t2 = t.loc[np.array([1, 4, 2])]
            assert_col_equal(t2["a"], [1, 4, 2])
        assert_col_equal(t2["a"], [1, 4, 2])
        t2 = t.loc[self.make_val(3) : self.make_val(5)]  # range search
        assert_col_equal(t2["a"], [3, 4, 5])
        t2 = t.loc.with_index("b")[5.0:7.0]
        assert_col_equal(t2["b"], [5.1, 6.2, 7.0])
        # search by sorted index
        t2 = t.iloc[0:2]  # two smallest rows by column 'a'
        assert_col_equal(t2["a"], [1, 2])
        t2 = t.iloc.with_index("b")[2:]  # exclude two smallest rows in column 'b'
        assert_col_equal(t2["b"], [5.1, 6.2, 7.0])

        for t2 in (t.loc[:], t.iloc[:]):
            assert_col_equal(t2["a"], [1, 2, 3, 4, 5])

    def test_table_loc_indices(self, main_col, table_types, engine):
        self._setup(main_col, table_types)
        t = self.t

        t.add_index("a", engine=engine)
        t.add_index("b", engine=engine)

        t2 = t.loc_indices[self.make_val(3)]  # single label, with primary key 'a'
        assert t2 == 2

        # list search
        t2 = t.loc_indices[[self.make_val(1), self.make_val(4), self.make_val(2)]]
        for i, p in zip(t2, [1, 4, 2]):  # same order as input list
            assert i == p - 1

    def test_invalid_search(self, main_col, table_types, engine):
        # using .loc and .loc_indices with a value not present should raise an exception
        self._setup(main_col, table_types)
        t = self.t

        t.add_index("a")
        with pytest.raises(KeyError):
            t.loc[self.make_val(6)]
        with pytest.raises(KeyError):
            t.loc_indices[self.make_val(6)]

    def test_copy_index_references(self, main_col, table_types, engine):
        # check against a bug in which indices were given an incorrect
        # column reference when copied
        self._setup(main_col, table_types)
        t = self.t

        t.add_index("a")
        t.add_index("b")
        t2 = t.copy()
        assert t2.indices["a"].columns[0] is t2["a"]
        assert t2.indices["b"].columns[0] is t2["b"]

    def test_unique_index(self, main_col, table_types, engine):
        self._setup(main_col, table_types)
        t = self.t

        t.add_index("a", engine=engine, unique=True)
        assert np.all(t.indices["a"].sorted_data() == [0, 1, 2, 3, 4])

        if self.mutable:
            with pytest.raises(ValueError):
                t.add_row((5.0, "9", 5))

    def test_copy_indexed_table(self, table_types):
        self._setup(_col, table_types)
        t = self.t
        t.add_index("a")
        t.add_index(["a", "b"])
        for tp in (self._table_type(t), t.copy()):
            assert len(t.indices) == len(tp.indices)
            for index, indexp in zip(t.indices, tp.indices):
                assert np.all(index.data.data == indexp.data.data)
                assert index.data.data.colnames == indexp.data.data.colnames

    def test_updating_row_byindex(self, main_col, table_types, engine):
        self._setup(main_col, table_types)
        t = Table(
            [["a", "b", "c", "d"], [2, 3, 4, 5], [3, 4, 5, 6]],
            names=("a", "b", "c"),
            meta={"name": "first table"},
        )

        t.add_index("a", engine=engine)
        t.add_index("b", engine=engine)

        t.loc["c"] = ["g", 40, 50]  # single label, with primary key 'a'
        t2 = t[2]
        assert list(t2) == ["g", 40, 50]

        # list search
        t.loc[["a", "d", "b"]] = [["a", 20, 30], ["d", 50, 60], ["b", 30, 40]]
        t2 = [["a", 20, 30], ["d", 50, 60], ["b", 30, 40]]
        for i, p in zip(t2, [1, 4, 2]):  # same order as input list
            assert list(t[p - 1]) == i

    def test_invalid_updates(self, main_col, table_types, engine):
        # using .loc and .loc_indices with a value not present should raise an exception
        self._setup(main_col, table_types)
        t = Table(
            [[1, 2, 3, 4], [2, 3, 4, 5], [3, 4, 5, 6]],
            names=("a", "b", "c"),
            meta={"name": "first table"},
        )

        t.add_index("a")
        with pytest.raises(ValueError):
            t.loc[3] = [[1, 2, 3]]
        with pytest.raises(ValueError):
            t.loc[[1, 4, 2]] = [[1, 2, 3], [4, 5, 6]]
        with pytest.raises(ValueError):
            t.loc[[1, 4, 2]] = [[1, 2, 3], [4, 5, 6], [2, 3]]
        with pytest.raises(ValueError):
            t.loc[[1, 4, 2]] = [[1, 2, 3], [4, 5], [2, 3]]


def test_get_index():
    a = [1, 4, 5, 2, 7, 4, 45]
    b = [2.0, 5.0, 8.2, 3.7, 4.3, 6.5, 3.3]
    t = Table([a, b], names=("a", "b"), meta={"name": "first table"})
    t.add_index(["a"])
    # Getting the values of index using names
    x1 = get_index(t, names=["a"])

    assert isinstance(x1, SlicedIndex)
    assert len(x1.columns) == 1
    assert len(x1.columns[0]) == 7
    assert x1.columns[0].info.name == "a"
    # Getting the vales of index using table_copy
    x2 = get_index(t, table_copy=t[["a"]])

    assert isinstance(x2, SlicedIndex)
    assert len(x2.columns) == 1
    assert len(x2.columns[0]) == 7
    assert x2.columns[0].info.name == "a"

    with pytest.raises(ValueError):
        get_index(t, names=["a"], table_copy=t[["a"]])
    with pytest.raises(ValueError):
        get_index(t, names=None, table_copy=None)


@pytest.mark.parametrize("table_type", [Table, QTable])
def test_index_loc_with_quantity(engine, table_type):
    t = table_type()
    t["a"] = [3, 1, 2] * u.m
    t["b"] = [1, 2, 3]
    t.add_index("a", engine=engine)

    unit = u.m if table_type is QTable else 1
    assert tuple(t.loc[1 * unit]) == (1 * unit, 2)
    assert np.all(t.loc_indices[[1 * unit, 3 * unit]] == [1, 0])
    assert tuple(t.iloc[1]) == (2 * unit, 3)
    for loc in (t.loc, t.iloc):
        t_loc = loc[:]
        assert len(t_loc) == 3
        assert np.all(t_loc["a"] == [1, 2, 3] * unit)
        assert np.all(t_loc["b"] == [2, 3, 1])


def test_index_loc_with_string(engine):
    t = Table()
    t["a"] = ["z", "a", "m"]
    t["b"] = [1, 2, 3]
    t.add_index("a", engine=engine)

    assert tuple(t.loc["a"]) == ("a", 2)
    assert np.all(t.loc_indices[["a", "z"]] == [1, 0])
    assert tuple(t.iloc[1]) == ("m", 3)
    for loc in (t.loc, t.iloc):
        t_loc = loc[:]
        assert len(t_loc) == 3
        assert np.all(t_loc["a"] == ["a", "m", "z"])
        assert np.all(t_loc["b"] == [2, 3, 1])


def test_table_index_time_warning(engine):
    # Make sure that no ERFA warnings are emitted when indexing a table by
    # a Time column with a non-default time scale
    tab = Table()
    tab["a"] = Time([1, 2, 3], format="jyear", scale="tai")
    tab["b"] = [4, 3, 2]
    with warnings.catch_warnings(record=True) as wlist:
        tab.add_index(("a", "b"), engine=engine)
    assert len(wlist) == 0


@pytest.mark.parametrize(
    "col",
    [
        Column(np.arange(50000, 50005)),
        np.arange(50000, 50005) * u.m,
        Time(np.arange(50000, 50005), format="mjd"),
    ],
)
def test_table_index_does_not_propagate_to_column_slices(col):
    # They lost contact to the parent table, so they should also not have
    # information on the indices; this helps prevent large memory usage if,
    # e.g., a large time column is turned into an object array; see gh-10688.
    tab = QTable()
    tab["t"] = col
    tab.add_index("t")
    t = tab["t"]
    assert t.info.indices
    tx = t[1:]
    assert not tx.info.indices
    tabx = tab[1:]
    t = tabx["t"]
    assert t.info.indices


def test_hstack_qtable_table():
    # Check in particular that indices are initialized or copied correctly
    # for a Column that is being converted to a Quantity.
    qtab = QTable([np.arange(5.0) * u.m], names=["s"])
    qtab.add_index("s")
    tab = Table([Column(np.arange(5.0), unit=u.s)], names=["t"])
    qstack = hstack([qtab, tab])
    assert qstack["t"].info.indices == []
    assert qstack.indices == []


def test_index_slice_exception():
    with pytest.raises(TypeError, match="index_slice must be tuple or slice"):
        SlicedIndex(None, None)


@pytest.fixture(scope="module")
def simple_table():
    """Simple table with an index on column 'a'."""
    t = Table()
    t["a"] = [3, 1, 2, 3]
    t["b"] = ["x", "y", "z", "w"]
    t.add_index("a")
    return t


@pytest.mark.parametrize("key", [None, "a"])
@pytest.mark.parametrize(
    "item,length,cls",
    [
        (slice(0, 0), 0, Table),
        ([], 0, Table),
        ([1], 1, Table),
        ([1, 3], 3, Table),
        (np.array([]), 0, Table),
        (np.array([1]), 1, Table),
        (3, 2, Table),  # scalar index with multiple rows
        (1, None, Row),  # scalar index with single row
    ],
)
def test_index_zero_slice_or_sequence_or_scalar(simple_table, key, item, length, cls):
    """Test that indexing with various types gives the expected result.

    Tests fix for #18037.
    """
    loc = simple_table.loc.with_index(key) if key is not None else simple_table.loc
    tloc = loc[item]
    assert isinstance(tloc, cls)
    assert tloc.colnames == simple_table.colnames

    rows = simple_table.loc_indices[item]
    if cls is Table:
        assert len(tloc) == length
        assert len(rows) == length


@pytest.mark.parametrize(
    "method,item",
    [
        ("loc", (2, 5)),
        ("iloc", 1),
        ("loc_indices", (2, 5)),
    ],
)
def test_index_id_item_deprecation_and_with_index(method, item):
    """t.loc/iloc/loc_indices[index_id, item] raises a deprecation warning.

    Also test that these methods
    """
    t = Table()
    t["a"] = [1, 2, 3]
    t["b"] = [4, 5, 6]
    t["c"] = ["x", "y", "z"]
    index_id = ("a", "b")
    t.add_index(index_id)
    prop = getattr(t, method)
    # Test calling like t.loc.with_index("a", "b") and t.loc.with_index(("a", "b")).
    out_call_1 = prop.with_index(*index_id)[item]
    out_call_2 = prop.with_index(index_id)[item]
    with pytest.warns(
        AstropyDeprecationWarning,
        match=r"Calling `Table.loc/iloc/loc_indices\[index_id, item\]`",
    ):
        out_depr = prop[index_id, item]
    assert type(out_depr) is type(out_call_1)
    assert out_depr == out_call_1
    assert type(out_call_1) is type(out_call_2)
    assert out_call_1 == out_call_2


def test_engine_type_error():
    t = Table()
    t["a"] = [1, 2]
    t["b"] = [3, 4]
    with pytest.raises(
        TypeError,
        match=r"engine must be an Engine class or instance, got 'b' instead.",
    ):
        t.add_index("a", "b")  # Easy mistake, too bad engine= is not keyword-only


@pytest.mark.parametrize(
    "masked",
    [pytest.param(False, id="raw-array"), pytest.param(True, id="masked array")],
)
def test_nd_columun_as_index(masked):
    # see https://github.com/astropy/astropy/issues/13292
    # and https://github.com/astropy/astropy/pull/16360
    t = Table()
    data = np.arange(0, 6)
    if masked:
        data = np.ma.masked_inside(data, 2, 4)
    t.add_column(data.reshape(3, -1), name="arr")
    with pytest.raises(
        ValueError, match="Multi-dimensional column 'arr' cannot be used as an index."
    ):
        t.add_index("arr")


@pytest.mark.parametrize("index_first", [True, False])
def test_slice_an_indexed_table(index_first):
    """Test slicing a table that is already indexed.

    Test of fix for https://github.com/astropy/astropy/issues/10732.

    #10732 is the case index_first=True, but also test slicing first (index_first=False)
    since we're at it.
    """
    t = Table()
    t["a"] = [9, 8, 7, 6, 5, 4, 3, 2, 1, 0]
    t["b"] = [0, 0, 0, 0, 0, 1, 1, 1, 1, 1]
    t["c"] = ["e", "f", "g", "h", "i", "j", "k", "a", "b", "c"]

    if index_first:
        t.add_index("a")
        t.add_index(["b", "c"])
        ts = t[::2]
    else:
        ts = t[::2]
        ts.add_index("a")
        ts.add_index(["b", "c"])

    assert ts.pformat() == [
        " a   b   c ",
        "--- --- ---",
        "  9   0   e",
        "  7   0   g",
        "  5   0   i",
        "  3   1   k",
        "  1   1   b",
    ]
    # Index access works
    assert str(ts.loc[5]).splitlines() == [
        " a   b   c ",
        "--- --- ---",
        "  5   0   i",
    ]

    # Remove row 2 (a==5), check index access still works
    ts.remove_row(2)
    assert ts.pformat() == [
        " a   b   c ",
        "--- --- ---",
        "  9   0   e",
        "  7   0   g",
        "  3   1   k",
        "  1   1   b",
    ]
    assert str(ts.loc[1]).splitlines() == [
        " a   b   c ",
        "--- --- ---",
        "  1   1   b",
    ]

    # Remove row 2 (now a==3), check index access still works
    ts.remove_row(2)
    assert ts.pformat() == [
        " a   b   c ",
        "--- --- ---",
        "  9   0   e",
        "  7   0   g",
        "  1   1   b",
    ]
    assert str(ts.loc[7]).splitlines() == [
        " a   b   c ",
        "--- --- ---",
        "  7   0   g",
    ]

    # Make sure primary index and secondary index look right (with original=True)
    assert str(ts.indices[0]).splitlines() == [
        "<SlicedIndex original=True index=<Index columns=('a',) data=<SortedArray length=3>",
        " a  rows",
        "--- ----",
        "  1    2",
        "  7    1",
        "  9    0>>",
    ]
    assert str(ts.indices[1]).splitlines() == [
        "<SlicedIndex original=True index=<Index columns=('b', 'c') data=<SortedArray length=3>",
        " b   c  rows",
        "--- --- ----",
        "  0   e    0",
        "  0   g    1",
        "  1   b    2>>",
    ]


def test_unique_indices_after_multicol_index_slice():
    """Test that table indices after slicing are correct.

    This tests code in Table._new_from_slice() that ensures uniqueness of table index
    objects when slicing (via slice, ndarray, list etc) a table with a multi-column
    index.
    """
    t = Table()
    t["a"] = [2, 3]
    t["b"] = [3, 5]
    t.add_index(["a", "b"])
    t2 = t[:1]
    assert len(t2.indices) == 1  # without fix would be 2, both with id ("a", "b").
    assert t2.indices[0].id == ("a", "b")



# Licensed under a 3-clause BSD style license - see LICENSE.rst

# ruff: noqa: FLY002

import textwrap
from importlib.util import module_from_spec, spec_from_file_location

import pytest

from astropy.tests.helper import _skip_docstring_tests_with_optimized_python


def test_composite_unit_definition(tmp_path):
    # Regression test for #17781 - error message was not informative
    module_path = tmp_path / "bad_module.py"
    module_path.write_text(
        textwrap.dedent("""
        from astropy import units as u
        from astropy.units.docgen import generate_unit_summary

        km_per_h = u.km / u.h

        __doc__ = generate_unit_summary(globals())
    """)
    )
    spec = spec_from_file_location("bad_module", module_path)
    module = module_from_spec(spec)

    with pytest.raises(
        TypeError, match=r"^'km_per_h' must be defined with 'def_unit\(\)'$"
    ):
        # emulate an import statement
        spec.loader.exec_module(module)


@_skip_docstring_tests_with_optimized_python
def test_unit_definition_module_summary_header():
    from astropy.units import si

    entries = si.__doc__.split("\n\n")
    assert len(entries) == 55
    assert entries[1] == "\n".join(
        (
            ".. list-table:: Available Units",
            "   :header-rows: 1",
            "   :widths: 10 20 20 20 1",
        )
    )
    assert entries[2] == "\n".join(
        (
            "   * - Unit",
            "     - Description",
            "     - Represents",
            "     - Aliases",
            "     - SI Prefixes",
        )
    )


@_skip_docstring_tests_with_optimized_python
def test_unit_definition_module_prefix_only_summary():
    from astropy.units import required_by_vounit

    entries = required_by_vounit.__doc__.split("\n\n")
    assert len(entries) == 6
    assert entries[3] == "\n".join(
        (
            "   * - Prefixes for ``solLum``",
            "     - Solar luminance prefixes",
            "     - :math:`\\mathrm{3.828 \\times 10^{26}\\,W}`",
            "     - ``L_sun``, ``Lsun``",
            "     - Only",
        )
    )
    assert entries[5] == "\n".join(
        (
            "   * - Prefixes for ``solRad``",
            "     - Solar radius prefixes",
            "     - :math:`\\mathrm{6.957 \\times 10^{8}\\,m}`",
            "     - ``R_sun``, ``Rsun``",
            "     - Only",
            "",
        )
    )


@_skip_docstring_tests_with_optimized_python
def test_unit_definition_module_normal_summary():
    from astropy.units import photometric

    entries = photometric.__doc__.split("\n\n")
    assert len(entries) == 10
    assert entries[5] == "\n".join(
        (
            "   * - ``AB``",
            "     - AB magnitude zero flux density (magnitude ``ABmag``).",
            "     - :math:`\\mathrm{3.6307805 \\times 10^{-20}\\,\\frac{erg}{Hz\\,s\\,cm^{2}}}`",
            "     - ``ABflux``",
            "     - No",
        )
    )
    assert entries[8] == "\n".join(
        (
            "   * - ``mgy``",
            "     - Maggies - a linear flux unit that is the flux for a mag=0 object.To tie this onto a specific calibrated unit system, the zero_point_flux equivalency should be used.",
            "     - ",
            "     - ``maggy``",
            "     - Yes",
        )
    )


@_skip_docstring_tests_with_optimized_python
def test_unit_definition_module_function_units_summary():
    from astropy.units.function import units

    entries = units.__doc__.split("\n\n")
    assert len(entries) == 14
    assert entries[8] == "\n".join(
        (
            ".. list-table:: Available Magnitude Units",
            "   :header-rows: 1",
            "   :widths: 10 50 10",
        )
    )
    assert entries[9] == "\n".join(
        (
            "   * - Unit",
            "     - Description",
            "     - Represents",
        )
    )
    assert entries[12] == "\n".join(
        (
            "   * - ``M_bol``",
            "     - Absolute bolometric magnitude: M_bol=0 corresponds to :math:`\\mathrm{3.0128 \\times 10^{28}\\,W}`",
            "     - mag(Bol)",
        )
    )



# Licensed under a 3-clause BSD style license - see LICENSE.rst
"""Astropy FLRW classes."""

__all__ = [  # noqa: RUF100, RUF022
    # base
    "FLRW",
    "FlatFLRWMixin",
    # lambdacdm
    "LambdaCDM",
    "FlatLambdaCDM",
    # w0cdm
    "wCDM",
    "FlatwCDM",
    # w0wacdm
    "w0waCDM",
    "Flatw0waCDM",
    # w0wzcdm
    "w0wzCDM",
    "Flatw0wzCDM",
    # wpwazpcdm
    "wpwaCDM",
    "FlatwpwaCDM",
]

from .base import FLRW, FlatFLRWMixin
from .lambdacdm import FlatLambdaCDM, LambdaCDM
from .w0cdm import FlatwCDM, wCDM
from .w0wacdm import Flatw0waCDM, w0waCDM
from .w0wzcdm import Flatw0wzCDM, w0wzCDM
from .wpwazpcdm import FlatwpwaCDM, wpwaCDM



# Licensed under a 3-clause BSD style license - see LICENSE.rst

"""
This module defines magnitude zero points and related photometric quantities.

The corresponding magnitudes are given in the description of each unit.
(the actual definitions are in `astropy.units.function.units`).

Both the units and magnitudes are available in (and should be used
through) the `astropy.units` namespace.

"""

import numpy as np

from astropy.constants.si import L_bol0

from . import astrophys, cgs, si
from .core import def_unit
from .docgen import generate_dunder_all, generate_unit_summary

__all__ = []  #  Units are added at the end

_ns = globals()

def_unit(
    ["Bol", "L_bol"],
    L_bol0,
    namespace=_ns,
    prefixes=False,
    doc=(
        "Luminosity corresponding to absolute bolometric magnitude zero "
        "(magnitude ``M_bol``)."
    ),
)
def_unit(
    ["bol", "f_bol"],
    L_bol0 / (4 * np.pi * (10.0 * astrophys.pc) ** 2),
    namespace=_ns,
    prefixes=False,
    doc=(
        "Irradiance corresponding to apparent bolometric magnitude zero "
        "(magnitude ``m_bol``)."
    ),
)
def_unit(
    ["AB", "ABflux"],
    10.0 ** (48.6 / -2.5) * cgs.erg * cgs.cm**-2 / si.s / si.Hz,
    namespace=_ns,
    prefixes=False,
    doc="AB magnitude zero flux density (magnitude ``ABmag``).",
)
def_unit(
    ["ST", "STflux"],
    10.0 ** (21.1 / -2.5) * cgs.erg * cgs.cm**-2 / si.s / si.AA,
    namespace=_ns,
    prefixes=False,
    doc="ST magnitude zero flux density (magnitude ``STmag``).",
)
def_unit(
    ["mgy", "maggy"],
    namespace=_ns,
    prefixes=[(["n"], ["nano"], 1e-9)],
    doc=(
        "Maggies - a linear flux unit that is the flux for a mag=0 object."
        "To tie this onto a specific calibrated unit system, the "
        "zero_point_flux equivalency should be used."
    ),
)


###########################################################################
# ALL & DOCSTRING

__all__ += generate_dunder_all(globals())  # noqa: PLE0605


if __doc__ is not None:
    # This generates a docstring for this module that describes all of the
    # standard units defined here.
    __doc__ += generate_unit_summary(globals())



# Licensed under a 3-clause BSD style license - see LICENSE.rst
"""
This module provides functionality to read CSV files into Astropy Tables using PyArrow.
"""

import datetime
import io
import os
from typing import TYPE_CHECKING, BinaryIO, Literal, Union

import numpy as np
from numpy.typing import DTypeLike, NDArray

from astropy.utils.compat.optional_deps import HAS_PYARROW
from astropy.utils.data import get_readable_fileobj

if TYPE_CHECKING:
    import pyarrow
    import pyarrow.csv

    from astropy.table import Table

__all__ = ["convert_pa_table_to_astropy_table", "read_csv"]


def read_csv(
    input_file: os.PathLike[str] | str | BinaryIO,
    *,
    delimiter: str = ",",
    quotechar: str | Literal[False] = '"',
    doublequote: bool = True,
    escapechar: str | bool = False,
    header_start: int | None = 0,
    data_start: int | None = None,
    names: list[str] | None = None,
    include_names: list[str] | None = None,
    dtypes: dict[str, DTypeLike] | None = None,
    comment: str | None = None,
    null_values: list[str] | None = None,
    encoding: str = "utf-8",
    newlines_in_values: bool = False,
    timestamp_parsers: list[str] | None = None,
) -> "Table":
    """Read a CSV file into an astropy Table using PyArrow.

    This function allows highly performant reading of text CSV files into an astropy
    ``Table`` using `PyArrow <https://arrow.apache.org/docs/python/csv.html>`_. The
    best performance is achieved for files with only numeric data types, but even for
    files with mixed data types, the performance is still better than the the standard
    ``astropy.io.ascii`` fast CSV reader.

    By default, empty values (zero-length string "") in the CSV file are read as masked
    values in the Table. This can be changed by using the ``null_values`` parameter to
    specify a list of strings to interpret as null (masked) values.

    Entirely empty lines in the CSV file are ignored.

    Columns consisting of only string values ``True`` and ``False`` are parsed as
    boolean data.

    Columns with ISO 8601 date/time strings are parsed as shown below:
    - ``12:13:14.123456``: ``object[datetime.time]``
    - ``2025-01-01``: ``np.datetime64[D]``
    - ``2025-01-01T01:02:03``: ``np.datetime64[s]``
    - ``2025-01-01T01:02:03.123456``: ``np.datetime64[ns]``

    Support for ignoring comment lines in the CSV file is provided by the ``comment``
    parameter. If this is set to a string, any line starting with optional whitespace
    and then this string is ignored. This is done by reading the entire file and
    scanning for comment lines. If the comment lines are all at the beginning of the
    file and both ``header_start`` and ``data_start`` are not specified, then the file
    is read efficiently by setting ``header_start`` to the first line after the
    comments. Otherwise the entire file is read into memory and the comment lines are
    removed before passing to the PyArrow CSV reader. Any values of ``header_start`` and
    ``data_start`` apply to the lines counts *after* the comment lines have been
    removed.

    Parameters
    ----------
    input_file : str, path-like, or file-like
        File path or binary file-like object to read from.
    delimiter : 1-character str, optional (default ",")
        Character delimiting individual cells in the CSV data.
    quotechar : 1-character str or False, optional (default '"')
        Character used optionally for quoting CSV values (`False` if quoting is not
        allowed).
    doublequote : bool, optional (default `True`)
        Whether two quotes in a quoted CSV value denote a single quote in the data.
    escapechar : 1-character str or `False`, optional (default `False`)
        Character used optionally for escaping special characters (`False` if escaping
        is not allowed).
    header_start : int, None, optional (default 0)
        Line index for the header line with column names. If `None` this implies that
        there is no header line and the column names are taken from ``names`` or
        generated automatically ("f0", "f1", ...).
    data_start : int, None, optional (default None)
        Line index for the start of data. If `None`, then data starts one line after the
        header line, or on the first line if there is no header.
    names : list, None, optional (default None)
        List of names for input data columns when there is no header line. If supplied,
        then ``header_start`` must be `None`.
    include_names : list, None, optional (default None)
        List of column names to include in output. If `None`, all columns are included.
    dtypes : dict[str, Any], None, optional (default None)
        If provided, this is a dictionary of data types for output columns. Each key is
        a column name and the value is either a PyArrow data type or a data type
        specifier that is accepted as an argument to `numpy.dtype`. Examples include
        ``pyarrow.Int32()``, ``pyarrow.time32("s")``, ``int``, ``np.float32``,
        ``np.dtype('f4')`` or ``"float32"``. Default is to infer the data types.
    comment : 1-character str or None, optional (default None)
        Character used to indicate the start of a comment. Any line starting with
        optional whitespace and then this character is ignored. Using this option will
        cause the parser to be slower and potentially use more memory as it uses Python
        code to strip comments.

    Other Parameters
    ----------------
    null_values : list, optional (default None)
        List of strings to interpret as null values. By default, only empty strings are
        considered as null values (equivalent to ``null_values=[""]``). Set to ``[]`` to
        disable null value handling.
    encoding: str, optional (default 'utf-8')
        Encoding of the input data.
    newlines_in_values: bool, optional (default False)
        Whether newline characters are allowed in CSV values. Setting this to True
        reduces the performance of multi-threaded CSV reading.
    timestamp_parsers: list, optional
        A sequence of strptime()-compatible format strings, tried in order when
        attempting to infer or convert timestamp values. The default is the special
        value ``pyarrow.csv.ISO8601`` uses the optimized internal ISO8601 parser.

    Returns
    -------
    astropy.table.Table
        An astropy Table containing the data from the CSV file.
    """
    check_has_pyarrow()
    from pyarrow import csv

    parse_options = csv.ParseOptions(
        delimiter=delimiter,
        quote_char=quotechar,
        double_quote=doublequote,
        escape_char=escapechar,
        newlines_in_values=newlines_in_values,
    )

    if comment is not None:
        input_file_stripped, header_start_new = strip_comment_lines(
            input_file, comment, encoding, header_start, data_start
        )
        if header_start_new is not None:
            header_start = header_start_new
        if input_file_stripped is not None:
            input_file = input_file_stripped

    read_options = get_read_options(header_start, data_start, names, encoding)
    convert_options = get_convert_options(
        include_names, dtypes, null_values, timestamp_parsers
    )

    with get_readable_fileobj(input_file, encoding="binary") as f:
        table_pa = csv.read_csv(
            f,
            parse_options=parse_options,
            read_options=read_options,
            convert_options=convert_options,
        )
    return convert_pa_table_to_astropy_table(table_pa)


def check_has_pyarrow():
    """Check for pyarrow and raise an informative exception if not available.

    Raises
    ------
    ModuleNotFoundError
        If pyarrow is not installed.
    """
    if not HAS_PYARROW:
        raise ModuleNotFoundError(
            "pyarrow is required for astropy.io.misc.pyarrow.csv functions"
        )


def convert_pa_string_array_to_numpy(arr: "pyarrow.Array") -> NDArray:
    """
    Convert a PyArrow string array to a NumPy array.

    Parameters
    ----------
    arr : pyarrow.Array
        The PyArrow string array to be converted.

    Returns
    -------
    np.ndarray
        The converted NumPy array.
    """
    check_has_pyarrow()
    import pyarrow as pa
    import pyarrow.compute as pc

    # This implementation is faster than these alternatives:
    # >>> arr.to_numpy().astype(str)
    # >>> np.array(arr.to_pylist(), dtype='str')
    max_length = pa.compute.max(pc.utf8_length(arr))
    out = np.empty(len(arr), dtype=f"U{max_length}")
    out[:] = arr.to_numpy()
    return out


def pyarrow_zero(data_type: "pyarrow.DataType"):
    """
    Return a "zero" value for the given PyArrow data type.

    This function provides a default value corresponding to the specified
    PyArrow data type. The returned value is intended to represent a
    neutral or "zero" equivalent for the type.

    Parameters
    ----------
    data_type : pa.DataType
        The PyArrow data type for which to determine the zero value.

    Returns
    -------
    Any
        A zero-equivalent value for the specified data type. The type of the
        returned value depends on the input data type:
        - Integer or floating types: 0
        - Boolean type: False
        - String or large string types: ""
        - Binary or large binary types: b""
        - Date type: datetime.date(1970, 1, 1)
        - Time type: datetime.time(0, 0, 0)
        - Timestamp type: datetime.datetime(1970, 1, 1)
        - Null type: None
        - Decimal type: decimal.Decimal(0)

    Raises
    ------
    NotImplementedError
        If no zero value is defined for the given data type.
    """
    check_has_pyarrow()
    import pyarrow as pa

    if pa.types.is_integer(data_type) or pa.types.is_floating(data_type):
        return 0
    elif pa.types.is_boolean(data_type):
        return False
    elif pa.types.is_string(data_type) or pa.types.is_large_string(data_type):
        return ""
    elif pa.types.is_binary(data_type) or pa.types.is_large_binary(data_type):
        return b""
    elif pa.types.is_date(data_type):
        return datetime.date(1970, 1, 1)
    elif pa.types.is_time(data_type):
        return datetime.time(0, 0, 0)
    elif pa.types.is_timestamp(data_type):
        return datetime.datetime(1970, 1, 1)
    elif pa.types.is_null(data_type):
        return None
    elif pa.types.is_decimal(data_type):
        import decimal

        return decimal.Decimal(0)
    else:
        raise NotImplementedError(f"No zero value defined for type: {data_type}")


def convert_pa_array_to_numpy(
    arr: Union["pyarrow.Array", "pyarrow.ChunkedArray"],
) -> NDArray:
    """
    Convert a PyArrow array to a NumPy array.

    Parameters
    ----------
    arr : pyarrow.Array, pyarrow.ChunkedArray
        The PyArrow array to be converted.

    Returns
    -------
    np.ndarray or np.ma.MaskedArray
        The converted NumPy array. If the input array contains null values, a masked
        array is returned with nulls masked out.

    Notes
    -----
    - If the input array does not contain null values, the result is an ndarray.
    - If the input array contains null values, they are replaced with a fill value
      (equivalent to ``np.zeros((), dtype)``)  before conversion, and a masked array is
      returned with the null positions masked.
    - The current implementation supports only int, float, bool, and string types.
    """
    check_has_pyarrow()
    import pyarrow as pa

    # Validate input type and set the fill value
    is_string = pa.types.is_string(arr.type)

    if arr.null_count == 0:
        # No nulls, just return an ndarray view of the pyarray
        out = convert_pa_string_array_to_numpy(arr) if is_string else arr.to_numpy()
    else:
        mask = arr.is_null().to_numpy()
        # Fill nulls in `arr` with zero. We do not know of a zero-copy fill for
        # pyarrow arrays with nulls, so we need to copy the data.
        arr = arr.fill_null(pyarrow_zero(arr.type))
        data = convert_pa_string_array_to_numpy(arr) if is_string else arr.to_numpy()
        out = np.ma.array(data, mask=mask, copy=False)

    return out


def convert_pa_table_to_astropy_table(table_pa) -> "Table":
    """
    Convert a PyArrow Table to an Astropy Table.

    Parameters
    ----------
    table_pa : ``pyarrow.Table``
        The PyArrow Table to be converted.

    Returns
    -------
    astropy.table.Table
        Converted astropy Table.
    """
    from astropy.table import Table

    columns = {
        name: convert_pa_array_to_numpy(col)
        for name, col in zip(table_pa.column_names, table_pa.itercolumns())
    }
    return Table(columns, copy=False)


def strip_comment_lines(
    input_file: os.PathLike | str | BinaryIO,
    comment: str,
    encoding: str,
    header_start: int | None,
    data_start: int | None,
) -> tuple[io.BytesIO | None, int | None]:
    """
    Handle specified comment string when reading ``input_file``.

    This function has two modes of operation:

    1. If ``header_start`` is `None` or 0, then read the file line by line, looking for
       lines that start with the specified comment string. If all comment lines are at
       the beginning of the file (a common case), then the lines can be efficiently
       skipped within the pyarrow CSV reader. In this case the return value is (`None`,
       header_start after comments).
    2. Otherwise, the function reads the entire file into memory in a BytesIO object,
       removing all lines that start with the specified comment string along the way. In
       this case the return value is (BytesIO object, `None`).

    Checking for the comment string ignores leading whitespace.

    If ``input_file`` is a file path, it will be opened in binary read mode. The
    ``comment`` string is encoded to bytes using the provided encoding for comparison
    with the file content.

    Parameters
    ----------
    input_file : os.PathLike, str, or BinaryIO
        The input file path or file-like object to read from.
    comment : str
        The comment string that identifies lines to be removed.
    encoding : str
        The encoding to use for the input file.
    header_start : int or None
        The line index where the header starts. See ``read_csv`` for details.
    data_start : int or None
        The line index where the data starts. See ``read_csv`` for details.

    Returns
    -------
    output_file : io.BytesIO | None
        A BytesIO object containing the filtered content with comment lines removed, or
        `None` if all comment lines are at the beginning of the file.
    header_start : int | None
        The line index after the last comment line, or `None` if not all comment lines
        are at the beginning of the file.
    """
    comment_encode = comment.encode(encoding)

    with get_readable_fileobj(input_file, encoding="binary") as f:
        if header_start in (None, 0) and data_start is None:
            idx_last_comment = -1
            for idx, line in enumerate(f):
                if line.lstrip().startswith(comment_encode):
                    if idx - idx_last_comment == 1:
                        idx_last_comment = idx
                    else:
                        # Gap between comment lines, need to reset input file handle and
                        # break out to the logic below.
                        f.seek(0)
                        break
            else:
                f.seek(0)
                return None, idx_last_comment + 1

        # If we get here, we need to read the whole file and remove comment lines and
        # write into an output BytesIO file.
        output = io.BytesIO()
        for idx, line in enumerate(f):
            if not line.lstrip().startswith(comment_encode):
                output.write(line)

        # Set up for the consumer to read it.
        output.seek(0)

    return output, None


def get_convert_options(
    include_names: list[str] | None,
    dtypes: dict[str, DTypeLike] | None,
    null_values: list[str] | None,
    timestamp_parsers: list[str] | None,
) -> "pyarrow.csv.ConvertOptions":
    """
    Generate PyArrow CSV convert options.

    Parameters
    ----------
    include_names : list or None
        List of column names to include in the conversion. If None, all columns are
        included.
    dtypes : dict or None
        Dictionary mapping column names to their respective data types. If None, default
        data types are used.
    null_values : list or None, optional (default None)
        List of strings to interpret as null values. By default, only empty strings are
        considered as null values.
    timestamp_parsers : list or None
        A sequence of strptime()-compatible format strings, tried in order when
        attempting to infer or convert timestamp values. The default is the special
        value ``pyarrow.csv.ISO8601`` uses the optimized internal ISO8601 parser.

    Returns
    -------
    pyarrow.csv.ConvertOptions
        PyArrow CSV ConvertOptions object configured with the specified column names and
        data types.
    """
    check_has_pyarrow()
    import pyarrow as pa
    from pyarrow import csv

    convert_options = csv.ConvertOptions()
    convert_options.strings_can_be_null = True
    convert_options.null_values = [""] if null_values is None else null_values

    if include_names is not None:
        convert_options.include_columns = include_names
    if dtypes is not None:
        convert_options.column_types = {
            colname: (
                dtype if isinstance(dtype, pa.DataType) else pa.from_numpy_dtype(dtype)
            )
            for colname, dtype in dtypes.items()
        }
    if timestamp_parsers is not None:
        convert_options.timestamp_parsers = timestamp_parsers
    return convert_options


def get_read_options(
    header_start: int | None,
    data_start: int | None,
    names: list[str] | None,
    encoding: str | None,
) -> "pyarrow.csv.ReadOptions":
    """
    Generate PyArrow CSV read options.

    Parameters
    ----------
    header_start : int or None
        The row index where the header starts. If None, no header is present.
    data_start : int or None
        The row index where the data starts. If None, data starts immediately after the
        header.
    names : list or None
        A list of column names. If None, column names will be autogenerated if no header
        is present.
    encoding : str or None
        The encoding of the CSV file. If None, the default encoding is used.

    Returns
    -------
    pyarrow.csv.ReadOptions
        The configured read options for reading the CSV file.
    """
    check_has_pyarrow()
    from pyarrow import csv

    if header_start is not None and names is not None:
        raise ValueError("cannot specify `names` unless `header_start=None`")

    read_options = csv.ReadOptions()

    # No header: header_start = None, data_start = None or int
    #   Set column_names to an empty list and autogenerate_column_names to True.
    if header_start is None:
        read_options.skip_rows_after_names = 0 if data_start is None else data_start
        if names is None:
            # With these two in combination, column names are generated automatically
            # as "f0", "f1", ...
            read_options.column_names = []
            read_options.autogenerate_column_names = True
        else:
            # Explicitly set column names.
            read_options.column_names = names
            read_options.autogenerate_column_names = False
    else:
        # Header present, header_start is not None
        if data_start is None:
            data_start = header_start + 1
        read_options.skip_rows = header_start
        read_options.skip_rows_after_names = data_start - header_start - 1

    if encoding is not None:
        read_options.encoding = encoding

    return read_options


def register_pyarrow_csv_table():
    """
    Register pyarrow.csv with Unified I/O as a Table reader.
    """
    from astropy.io import registry as io_registry
    from astropy.table import Table

    io_registry.register_reader("pyarrow.csv", Table, read_csv)



import numpy as np
import pytest
from numpy.testing import assert_allclose, assert_equal

from astropy import units as u
from astropy.stats.circstats import (
    _length,
    circcorrcoef,
    circmean,
    circmoment,
    circvar,
    rayleightest,
    vonmisesmle,
    vtest,
)
from astropy.utils.compat.optional_deps import HAS_SCIPY


def test__length():
    # testing against R CircStats package
    # Ref. [1] pages 6 and 125
    weights = np.array([12, 1, 6, 1, 2, 1, 1])
    answer = 0.766282
    data = np.array([0, 3.6, 36, 72, 108, 169.2, 324]) * u.deg
    assert_allclose(answer, _length(data, weights=weights), atol=1e-4)


def test_circmean():
    # testing against R CircStats package
    # Ref[1], page 23
    data = np.array([51, 67, 40, 109, 31, 358]) * u.deg
    answer = 48.63 * u.deg
    assert_equal(answer, np.around(circmean(data), 2))


@pytest.mark.skipif(not HAS_SCIPY, reason="requires scipy")
def test_circmean_against_scipy():
    import scipy.stats

    # testing against scipy.stats.circmean function
    # the data is the same as the test before, but in radians
    data = np.array(
        [0.89011792, 1.1693706, 0.6981317, 1.90240888, 0.54105207, 6.24827872]
    )
    answer = scipy.stats.circmean(data)
    assert_equal(np.around(answer, 2), np.around(circmean(data), 2))


def test_circvar():
    # testing against R CircStats package
    # Ref[1], page 23
    data = np.array([51, 67, 40, 109, 31, 358]) * u.deg
    answer = 0.1635635
    assert_allclose(answer, circvar(data), atol=1e-4)


def test_circmoment():
    # testing against R CircStats package
    # Ref[1], page 23
    data = np.array([51, 67, 40, 109, 31, 358]) * u.deg
    # 2nd, 3rd, and 4th moments
    # this is the answer given in Ref[1] in radians
    answer = np.array([1.588121, 1.963919, 2.685556])
    answer = np.around(np.rad2deg(answer) * u.deg, 4)

    result = (
        np.around(circmoment(data, p=2)[0], 4),
        np.around(circmoment(data, p=3)[0], 4),
        np.around(circmoment(data, p=4)[0], 4),
    )

    assert_equal(answer[0], result[0])
    assert_equal(answer[1], result[1])
    assert_equal(answer[2], result[2])
    # testing lengths
    answer = np.array([0.4800428, 0.236541, 0.2255761])
    assert_allclose(
        answer,
        (circmoment(data, p=2)[1], circmoment(data, p=3)[1], circmoment(data, p=4)[1]),
        atol=1e-4,
    )


def test_circcorrcoef():
    # testing against R CircStats package
    # Ref[1], page 180

    # fmt: off
    alpha = np.array(
        [
            356, 97, 211, 232, 343, 292, 157, 302, 335, 302, 324,
            85, 324, 340, 157, 238, 254, 146, 232, 122, 329,
        ]
    ) * u.deg

    beta = np.array(
        [
            119, 162, 221, 259, 270, 29, 97, 292, 40, 313, 94, 45,
            47, 108, 221, 270, 119, 248, 270, 45, 23,
        ]
    ) * u.deg
    # fmt: on

    answer = 0.2704648
    assert_allclose(answer, circcorrcoef(alpha, beta), atol=1e-4)


def test_rayleightest():
    # testing against R CircStats package
    data = np.array([190.18, 175.48, 155.95, 217.83, 156.36]) * u.deg
    # answer was obtained through R CircStats function r.test(x)
    answer = (0.00640418, 0.9202565)
    result = (rayleightest(data), _length(data))
    assert_allclose(answer[0], result[0], atol=1e-4)
    assert_allclose(answer[1], result[1], atol=1e-4)


@pytest.mark.skipif(not HAS_SCIPY, reason="requires scipy")
def test_vtest():
    # testing against R CircStats package
    data = np.array([190.18, 175.48, 155.95, 217.83, 156.36]) * u.deg
    # answer was obtained through R CircStats function v0.test(x)
    answer = 0.9994725
    assert_allclose(answer, vtest(data), atol=1e-5)


def test_vonmisesmle():
    # testing against R CircStats package
    # testing non-Quantity

    # fmt: off
    data = np.array(
        [
            3.3699057, 4.0411630, 0.5014477, 2.6223103, 3.7336524,
            1.8136389, 4.1566039, 2.7806317, 2.4672173, 2.8493644,
        ]
    )
    # fmt: on

    # answer was obtained through R CircStats function vm.ml(x)
    answer = (3.006514, 1.474132)
    assert_allclose(answer[0], vonmisesmle(data)[0], atol=1e-5)
    assert_allclose(answer[1], vonmisesmle(data)[1], atol=1e-5)

    # testing with Quantity
    data = np.rad2deg(data) * u.deg
    answer = np.rad2deg(3.006514) * u.deg
    assert_equal(np.around(answer, 3), np.around(vonmisesmle(data)[0], 3))

    # testing for weighted vonmisesmle
    data = np.array(
        [
            np.pi / 2,
            np.pi,
            np.pi / 2,
        ]
    )  # this data has twice more np.pi/2 than np.pi
    # get answer using astropy vonmisesmle to test
    answer = vonmisesmle(data)
    data_to_weigh = np.array([np.pi / 2, np.pi])
    weights = [2, 1]
    assert_allclose(
        answer[0], vonmisesmle(data_to_weigh, weights=weights)[0], atol=1e-5
    )
    assert_allclose(
        answer[1], vonmisesmle(data_to_weigh, weights=weights)[1], atol=1e-5
    )

    # testing for axis argument (stacking the data from the first test)
    data = np.array(
        [
            [
                3.3699057,
                4.0411630,
                0.5014477,
                2.6223103,
                3.7336524,
                1.8136389,
                4.1566039,
                2.7806317,
                2.4672173,
                2.8493644,
            ],
            [
                3.3699057,
                4.0411630,
                0.5014477,
                2.6223103,
                3.7336524,
                1.8136389,
                4.1566039,
                2.7806317,
                2.4672173,
                2.8493644,
            ],
        ]
    )
    # answer should be duplicated
    answer = (np.array([3.006514, 3.006514]), np.array([1.474132, 1.474132]))
    assert_allclose(answer[0], vonmisesmle(data, axis=1)[0], atol=1e-5)
    assert_allclose(answer[1], vonmisesmle(data, axis=1)[1], atol=1e-5)

    # same test for Quantity
    data = np.rad2deg(data) * u.deg
    answer = (np.rad2deg(answer[0]) * u.deg, answer[1])
    assert_allclose(answer[0], vonmisesmle(data, axis=1)[0], atol=1e-5)
    assert_allclose(answer[1], vonmisesmle(data, axis=1)[1], atol=1e-5)



# Licensed under a 3-clause BSD style license

from pathlib import Path

from numpy import get_include as get_numpy_include
from setuptools import Extension

ROOT = Path(__file__).parent.relative_to(Path.cwd())


def get_extensions():
    sources = [
        str(ROOT / "cparser.pyx"),
        str(ROOT / "src" / "tokenizer.c"),
    ]
    ascii_ext = Extension(
        name="astropy.io.ascii.cparser",
        include_dirs=[get_numpy_include()],
        sources=sources,
    )
    return [ascii_ext]
